from spaceone.core.error import *


class ERROR_DOMAIN_STATE(ERROR_BASE):
    _message = 'Domain is disabled (domain_id = {domain_id})'


class ERROR_DOMAIN_AUTH_PLUGIN(ERROR_BASE):
    _message = 'Failed to get endpoint of version = {version}, domain_id = {domain_id}'


class ERROR_EXTERNAL_USER_NOT_ALLOWED_API_USER(ERROR_INVALID_ARGUMENT):
    _message = 'External user cannot be created with the API_USER type.'


class ERROR_NOT_ALLOWED_ROLE_TYPE(ERROR_INVALID_ARGUMENT):
    _message = 'Duplicate assignment of system roles and domain or project roles is not allowed.'


class ERROR_NOT_ALLOWED_EXTERNAL_AUTHENTICATION(ERROR_INVALID_ARGUMENT):
    _message = 'This domain does not allow external authentication.'


class ERROR_PLUGIN_IS_NOT_SET(ERROR_BASE):
    _message = "External auth plugin is not set. First, execute the 'change_auth_plugin' API to set the plugin."



import re
import logging

from spaceone.core.manager import BaseManager
from spaceone.identity.lib.cipher import PasswordCipher
from spaceone.identity.model.domain_owner_model import DomainOwner
from spaceone.identity.error.error_user import *

_LOGGER = logging.getLogger(__name__)


class DomainOwnerManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.domain_owner_model: DomainOwner = self.locator.get_model('DomainOwner')

    def create_owner(self, params):
        def _rollback(vo):
            _LOGGER.info(f'[create_owner._rollback] Delete domain owner : {vo["owner_id"]} ({vo["domain_id"]})')
            vo.delete()

        if params.get('password'):
            self._check_password_format(params['password'])
            hashed_pw = PasswordCipher().hashpw(params['password'])
            params['password'] = hashed_pw

        domain_owner: DomainOwner = self.domain_owner_model.create(params)

        self.transaction.add_rollback(_rollback, domain_owner)

        return domain_owner

    def update_owner(self, params):
        def _rollback(old_vo):
            _LOGGER.info(f'[update_owner._rollback] Revert domain owner : {old_vo["name"]} ({old_vo["domain_id"]})')
            domain_owner.update(old_vo)

        if params.get('password'):
            self._check_password_format(params['password'])
            hashed_pw = PasswordCipher().hashpw(params['password'])
            params['password'] = hashed_pw

        domain_owner: DomainOwner = self.domain_owner_model.get(owner_id=params['owner_id'], domain_id=params['domain_id'])

        self.transaction.add_rollback(_rollback, domain_owner.to_dict())

        return domain_owner.update(params)

    def delete_owner(self, domain_id, owner_id):
        domain_owner: DomainOwner = self.domain_owner_model.get(domain_id=domain_id, owner_id=owner_id)
        domain_owner.delete()

    def get_owner(self, domain_id, owner_id=None, only=None):
        if owner_id:
            return self.domain_owner_model.get(domain_id=domain_id, owner_id=owner_id, only=only)
        else:
            return self.domain_owner_model.get(domain_id=domain_id, only=only)

    @staticmethod
    def _check_password_format(password):
        if len(password) < 8:
            raise ERROR_INCORRECT_PASSWORD_FORMAT(rule='At least 9 characters long.')
        elif not re.search("[a-z]", password):
            raise ERROR_INCORRECT_PASSWORD_FORMAT(rule='Contains at least one lowercase character')
        elif not re.search("[A-Z]", password):
            raise ERROR_INCORRECT_PASSWORD_FORMAT(rule='Contains at least one uppercase character')
        elif not re.search("[0-9]", password):
            raise ERROR_INCORRECT_PASSWORD_FORMAT(rule='Contains at least one number')



import logging

from spaceone.core import cache
from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector
from spaceone.identity.error.error_domain import *
from spaceone.identity.connector import AuthPluginConnector
from spaceone.identity.model.domain_model import Domain

_LOGGER = logging.getLogger(__name__)


class DomainManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.domain_model: Domain = self.locator.get_model('Domain')

    def create_domain(self, params):
        def _rollback(vo):
            _LOGGER.info(f'[create_domain._rollback] Delete domain : {vo.name} ({vo.domain_id})')
            vo.delete()

        domain_vo: Domain = self.domain_model.create(params)
        self.transaction.add_rollback(_rollback, domain_vo)

        if 'plugin_info' in params:
            domain_id = domain_vo.domain_id
            plugin_info = params['plugin_info']
            options = plugin_info.get('options', {})
            endpoint, updated_version = self.get_auth_plugin_endpoint(domain_id, plugin_info)

            if updated_version:
                params['plugin_info']['version'] = updated_version

            response = self.init_auth_plugin(endpoint, options)
            params['plugin_info']['metadata'] = response['metadata']

            secret_data = plugin_info.get('secret_data')

            if secret_data:
                schema = plugin_info.get('schema')
                secret_id = self._create_secret(domain_id, secret_data, schema)

                if secret_id:
                    params['plugin_info']['secret_id'] = secret_id
                    del params['plugin_info']['secret_data']

            return domain_vo.update({
                'plugin_info': params['plugin_info']
            })
        else:
            return domain_vo

    def update_domain(self, params):
        def _rollback(old_data):
            _LOGGER.info(f'[update_domain._rollback] Revert Data : {old_data["name"]} ({old_data["domain_id"]})')
            domain_vo.update(old_data)

        domain_vo: Domain = self.get_domain(params['domain_id'])
        self.transaction.add_rollback(_rollback, domain_vo.to_dict())

        return domain_vo.update(params)

    def change_auth_plugin(self, domain_id, plugin_info):
        """"
        Case 1. No plugin_info -> New plugin_info
        Case 2. plugin_info(1) -> New plugin_info
        """
        domain_vo: Domain = self.get_domain(domain_id)
        secret_id = None

        try:
            secret_id = domain_vo.plugin_info.secret_id
        except Exception as e:
            # No plugin_info or plugin_info without secret_id
            pass

        options = plugin_info.get('options', {})
        endpoint, updated_version = self.get_auth_plugin_endpoint(domain_id, plugin_info)

        if updated_version:
            plugin_info['version'] = updated_version

        response = self.init_auth_plugin(endpoint, options)
        plugin_info['metadata'] = response['metadata']

        secret_data = plugin_info.get('secret_data')

        if secret_data:
            schema = plugin_info.get('schema')

            if secret_id:
                self._update_secret_data(secret_id, domain_id, secret_data, schema)
            else:
                secret_id = self._create_secret(domain_id, secret_data, schema)

            if secret_id:
                plugin_info['secret_id'] = secret_id
                del plugin_info['secret_data']

        updated_domain_vo = domain_vo.update({'plugin_info': plugin_info})

        return updated_domain_vo

    def release_auth_plugin(self, domain_id):
        """ release plugin_info
        """
        domain_vo: Domain = self.get_domain(domain_id)

        # clean plugin_info
        # secret_id, if exist
        if domain_vo.plugin_info.secret_id:
            self._delete_secret(domain_vo.plugin_info.secret_id, domain_id)

        return domain_vo.update({'plugin_info': None})

    def update_domain_plugin(self, domain_id, version=None, options=None, upgrade_mode=None):
        """ Update plugin of domain
        If options exists, it should be complete content.
        """
        domain_vo: Domain = self.get_domain(domain_id)

        if domain_vo.plugin_info is None:
            raise ERROR_PLUGIN_IS_NOT_SET()

        plugin_info = domain_vo.plugin_info.to_dict()

        if version:
            plugin_info['version'] = version

        if options:
            plugin_info['options'] = options

        if upgrade_mode:
            plugin_info['upgrade_mode'] = upgrade_mode

        endpoint, updated_version = self.get_auth_plugin_endpoint(domain_id, plugin_info)

        if updated_version:
            plugin_info['version'] = updated_version

        response = self.init_auth_plugin(endpoint, plugin_info.get('options', {}))
        plugin_info['metadata'] = response['metadata']

        return domain_vo.update({'plugin_info': plugin_info})

    def verify_auth_plugin(self, domain_id):
        domain_vo: Domain = self.get_domain(domain_id)
        endpoint = self.get_auth_plugin_endpoint_by_vo(domain_vo)
        plugin_info = domain_vo.plugin_info.to_dict()

        auth_conn: AuthPluginConnector = self.locator.get_connector('AuthPluginConnector')
        auth_conn.initialize(endpoint)

        return auth_conn.verify(plugin_info['options'])

    def delete_domain(self, domain_id):
        domain_vo: Domain = self.get_domain(domain_id)
        domain_vo.delete()

        cache.delete_pattern(f'domain-state:{domain_id}')

    def enable_domain(self, domain_id):
        def _rollback(old_data):
            _LOGGER.info(f'[enable_domain._rollback] Revert Data : {old_data["name"]} ({old_data["domain_id"]})')
            domain_vo.update(old_data)

        domain_vo: Domain = self.get_domain(domain_id)

        if domain_vo.state != 'ENABLED':
            self.transaction.add_rollback(_rollback, domain_vo.to_dict())
            domain_vo.update({'state': 'ENABLED'})

            cache.delete_pattern(f'domain-state:{domain_id}')

        return domain_vo

    def disable_domain(self, domain_id):
        def _rollback(old_data):
            _LOGGER.info(f'[disable_domain._rollback] Revert Data : {old_data["name"]} ({old_data["domain_id"]})')
            domain_vo.update(old_data)

        domain_vo: Domain = self.get_domain(domain_id)

        if domain_vo.state != 'DISABLED':
            self.transaction.add_rollback(_rollback, domain_vo.to_dict())
            domain_vo.update({'state': 'DISABLED'})

            cache.delete_pattern(f'domain-state:{domain_id}')

        return domain_vo

    def get_domain(self, domain_id, only=None):
        return self.domain_model.get(domain_id=domain_id, only=only)

    def list_domains(self, query):
        return self.domain_model.query(**query)

    def stat_domains(self, query):
        return self.domain_model.stat(**query)

    def get_auth_plugin_endpoint_by_vo(self, domain_vo: Domain):
        plugin_info = domain_vo.plugin_info.to_dict()
        endpoint, updated_version = self.get_auth_plugin_endpoint(domain_vo.domain_id, plugin_info)

        if updated_version:
            _LOGGER.debug(f'[get_auth_plugin_endpoint_by_vo] upgrade plugin version: {plugin_info["version"]} -> {updated_version}')
            self.upgrade_auth_plugin_version(domain_vo, endpoint, updated_version)

        return endpoint

    def get_auth_plugin_endpoint(self, domain_id, plugin_info):
        plugin_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', service='plugin')
        response = plugin_connector.dispatch(
            'Plugin.get_plugin_endpoint',
            {
                'plugin_id': plugin_info['plugin_id'],
                'version': plugin_info.get('version'),
                'upgrade_mode': plugin_info.get('upgrade_mode', 'AUTO'),
                'domain_id': domain_id
            }
        )

        return response['endpoint'], response.get('updated_version')

    def upgrade_auth_plugin_version(self, domain_vo: Domain, endpoint, updated_version):
        plugin_info = domain_vo.plugin_info.to_dict()
        response = self.init_auth_plugin(endpoint, plugin_info.get('options', {}))
        plugin_info['version'] = updated_version
        plugin_info['metadata'] = response['metadata']
        domain_vo.update({'plugin_info': plugin_info})

    def init_auth_plugin(self, endpoint, options):
        auth_conn: AuthPluginConnector = self.locator.get_connector('AuthPluginConnector')
        auth_conn.initialize(endpoint)

        return auth_conn.init(options)

    def _create_secret(self, domain_id, secret_data, schema):
        secret_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', service='secret')
        params = {
            'name': f'{domain_id}-auth-plugin-credentials',
            'data': secret_data,
            'secret_type': 'CREDENTIALS',
            'schema': schema,
            'domain_id': domain_id
        }

        resp = secret_connector.dispatch('Secret.create', params)
        _LOGGER.debug(f'[_create_secret] {resp}')
        return resp.get('secret_id')

    def _delete_secret(self, secret_id, domain_id):
        secret_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', service='secret')
        params = {
            'secret_id': secret_id,
            'domain_id': domain_id
        }
        secret_connector.dispatch('Secret.delete', params)

    def _update_secret_data(self, secret_id, domain_id, secret_data, schema):
        secret_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', service='secret')
        params = {
            'secret_id': secret_id,
            'data': secret_data,
            'schema': schema,
            'domain_id': domain_id,
        }

        secret_connector.dispatch('Secret.update_data', params)



from mongoengine import *
from datetime import datetime
from spaceone.core.error import *
from spaceone.core.model.mongo_model import MongoModel


class PluginInfo(EmbeddedDocument):
    plugin_id = StringField(max_length=40)
    version = StringField(max_length=255)
    options = DictField(default={})
    metadata = DictField(default={})
    secret_id = StringField(max_length=40, null=True, default=None)
    schema = StringField(max_length=255, null=True, default=None)
    upgrade_mode = StringField(max_length=20, default='AUTO', choices=('AUTO', 'MANUAL'))

    def to_dict(self):
        return dict(self.to_mongo())


class Domain(MongoModel):
    domain_id = StringField(max_length=40, generate_id='domain', unique=True)
    name = StringField(max_length=255)
    state = StringField(max_length=20, default='ENABLED')
    plugin_info = EmbeddedDocumentField(PluginInfo, default=None, null=True)
    config = DictField()
    tags = DictField()
    created_at = DateTimeField(auto_now_add=True)
    deleted_at = DateTimeField(default=None, null=True)

    meta = {
        'updatable_fields': [
            'name',
            'state',
            'plugin_info',
            'config',
            'tags',
            'deleted_at'
        ],
        'minimal_fields': [
            'domain_id',
            'name',
            'state',
        ],
        'ordering': ['name'],
        'indexes': [
            # 'domain_id',
            'state',
        ]
    }

    @queryset_manager
    def objects(doc_cls, queryset):
        return queryset.filter(state__ne='DELETED')

    @classmethod
    def create(cls, data):
        domain_vos = cls.filter(name=data['name'])
        if domain_vos.count() > 0:
            raise ERROR_NOT_UNIQUE(key='name', value=data['name'])

        return super().create(data)

    def update(self, data):
        if 'name' in data:
            domain_vos = self.filter(name=data['name'], domain_id__ne=self.domain_id)
            if domain_vos.count() > 0:
                raise ERROR_NOT_UNIQUE(key='name', value=data['name'])

        return super().update(data)

    def delete(self):
        self.update({
            'state': 'DELETED',
            'deleted_at': datetime.utcnow()
        })



from spaceone.core.service import *
from spaceone.identity.manager import APIKeyManager, UserManager


@authentication_handler(exclude=['get'])
@authorization_handler(exclude=['get'])
@mutation_handler
@event_handler
class APIKeyService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.api_key_mgr: APIKeyManager = self.locator.get_manager('APIKeyManager')

    @transaction(append_meta={'authorization.scope': 'USER'})
    @check_required(['user_id', 'domain_id'])
    def create(self, params):
        """ Create api key

        Args:
            params (dict): {
                'user_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            api_key_vo (object)
        """

        user_id = params['user_id']
        domain_id = params['domain_id']

        # Check user is exists.
        user_mgr: UserManager = self.locator.get_manager('UserManager')
        user_vo = user_mgr.get_user(user_id=user_id, domain_id=domain_id)

        return self.api_key_mgr.create_api_key(user_vo, domain_id)

    @transaction(append_meta={'authorization.scope': 'USER'})
    @check_required(['api_key_id', 'domain_id'])
    def enable(self, params):
        """ Enable api key

        Args:
            params (dict): {
                'api_key_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            api_key_vo (object)
        """

        return self.api_key_mgr.enable_api_key(params['api_key_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'USER'})
    @check_required(['api_key_id', 'domain_id'])
    def disable(self, params):
        """ Disable api key

        Args:
            params (dict): {
                'api_key_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            api_key_vo (object)
        """

        return self.api_key_mgr.disable_api_key(params['api_key_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'USER'})
    @check_required(['api_key_id', 'domain_id'])
    def delete(self, params):
        """ Delete api key

        Args:
            params (dict): {
                'api_key_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        self.api_key_mgr.delete_api_key(params['api_key_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'USER'})
    @check_required(['api_key_id', 'domain_id'])
    def get(self, params):
        """ Get api key

        Args:
            params (dict): {
                'api_key_id': 'str',
                'domain_id': 'str',
                'only': 'list'
            }

        Returns:
            api_key_vo (object)
        """

        return self.api_key_mgr.get_api_key(params['api_key_id'], params['domain_id'], params.get('only'))

    @transaction(append_meta={'authorization.scope': 'USER'})
    @check_required(['domain_id'])
    @append_query_filter(['api_key_id', 'state', 'user_id', 'domain_id'])
    @append_keyword_filter(['api_key_id', 'user_id'])
    def list(self, params):
        """ List api keys

        Args:
            params (dict): {
                'api_key_id': 'str',
                'state': 'str',
                'user_id': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            results (list): 'list of api_key_vo'
            total_count (int)
        """

        return self.api_key_mgr.list_api_keys(params.get('query', {}))

    @transaction(append_meta={'authorization.scope': 'USER'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id'])
    @append_keyword_filter(['api_key_id', 'user_id'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
            }

        Returns:
            values (list): 'list of statistics data'
            total_count (int)
        """

        query = params.get('query', {})
        return self.api_key_mgr.stat_api_keys(query)



import logging
import os
from jinja2 import Environment, FileSystemLoader, select_autoescape

from spaceone.core import config, utils
from spaceone.core.manager import BaseManager
from spaceone.identity.connector.smtp_connector import SMTPConnector

_LOGGER = logging.getLogger(__name__)

TEMPLATE_PATH = os.path.join(os.path.dirname(__file__), f'../template')
JINJA_ENV = Environment(
    loader=FileSystemLoader(searchpath=TEMPLATE_PATH),
    autoescape=select_autoescape()
)

LANGUAGE_MAPPER = {
    'default': {
        'reset_password': 'Reset your password',
        'temp_password': 'Your password has been changed',
        'verify_email': 'Verify your notification email',
    },
    'ko': {
        'reset_password': '비밀번호 재설정 안내',
        'temp_password': '임시 비밀번호 발급 안내',
        'verify_email': '알림전용 이메일 계정 인증 안내',
    },
    'en': {
        'reset_password': 'Reset your password',
        'temp_password': 'Your password has been changed',
        'verify_email': 'Verify your notification email',
    },
    'ja': {
        'reset_password': 'パスワードリセットのご案内',
        'temp_password': '仮パスワード発行のご案内',
        'verify_email': '通知メールアカウント認証のご案内',
    }

}


class EmailManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.smtp_connector: SMTPConnector = self.locator.get_connector('SMTPConnector')

    def send_reset_password_email(self, user_id, email, reset_password_link, language):
        service_name = self._get_service_name()
        language_map_info = LANGUAGE_MAPPER.get(language, 'default')

        template = JINJA_ENV.get_template(f'reset_pwd_link_when_pw_forgotten_{language}.html')
        email_contents = template.render(user_name=user_id, reset_password_link=reset_password_link, service_name=service_name)
        subject = f'[{service_name}] {language_map_info["reset_password"]}'

        self.smtp_connector.send_email(email, subject, email_contents)

    def send_temporary_password_email(self, user_id, email, console_link, temp_password, language):
        service_name = self._get_service_name()
        language_map_info = LANGUAGE_MAPPER.get(language, 'default')

        template = JINJA_ENV.get_template(f'temp_pwd_when_pw_forgotten_{language}.html')
        email_contents = template.render(user_name=user_id, temp_password=temp_password, service_name=service_name, login_link=console_link)
        subject = f'[{service_name}] {language_map_info["temp_password"]}'

        self.smtp_connector.send_email(email, subject, email_contents)

    def send_reset_password_email_when_user_added(self, user_id, email, reset_password_link, language):
        service_name = self._get_service_name()
        language_map_info = LANGUAGE_MAPPER.get(language, 'default')

        template = JINJA_ENV.get_template(f'reset_pwd_link_when_user_added_{language}.html')
        email_contents = template.render(user_name=user_id, reset_password_link=reset_password_link, service_name=service_name)
        subject = f'[{service_name}] {language_map_info["reset_password"]}'

        self.smtp_connector.send_email(email, subject, email_contents)

    def send_temporary_password_email_when_user_added(self, user_id, email, console_link, temp_password, language):
        service_name = self._get_service_name()
        language_map_info = LANGUAGE_MAPPER.get(language, 'default')

        template = JINJA_ENV.get_template(f'temp_pwd_when_user_added_{language}.html')
        email_contents = template.render(user_name=user_id, temp_password=temp_password, service_name=service_name, login_link=console_link)
        subject = f'[{service_name}] {language_map_info["temp_password"]}'

        self.smtp_connector.send_email(email, subject, email_contents)

    def send_verification_email(self, user_id, email, verification_code, language):
        service_name = self._get_service_name()
        language_map_info = LANGUAGE_MAPPER.get(language, 'default')

        template = JINJA_ENV.get_template(f'verification_code_{language}.html')
        email_contents = template.render(user_name=user_id, verification_code=verification_code, service_name=service_name)
        subject = f'[{service_name}] {language_map_info["verify_email"]}'

        self.smtp_connector.send_email(email, subject, email_contents)

    @staticmethod
    def _get_service_name():
        return config.get_global('EMAIL_SERVICE_NAME')




import logging
from datetime import datetime

from spaceone.identity.error.error_authentication import *
from spaceone.identity.lib.cipher import PasswordCipher
from spaceone.identity.manager.token_manager import JWTManager
from spaceone.identity.manager import DomainOwnerManager
from spaceone.identity.model import DomainOwner

_LOGGER = logging.getLogger(__name__)


class DomainOwnerTokenManager(JWTManager):
    user: DomainOwner = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.domain_owner_mgr: DomainOwnerManager = self.locator.get_manager('DomainOwnerManager')

    def authenticate(self, user_id, domain_id, credentials):
        pw_to_check = self._parse_password(credentials)

        self.user = self.domain_owner_mgr.get_owner(owner_id=user_id, domain_id=domain_id)

        is_correct = PasswordCipher().checkpw(pw_to_check, self.user.password)
        _LOGGER.debug(f'[authenticate] is_correct: {is_correct}, pw_to_check: {pw_to_check}, hashed_pw: {self.user.password}')

        if is_correct:
            self.is_authenticated = True
        else:
            raise ERROR_AUTHENTICATION_FAILURE(user_id=self.user.owner_id)

    def issue_token(self, **kwargs):
        if self.is_authenticated is False:
            raise ERROR_NOT_AUTHENTICATED()

        # Issue token
        access_token = self.issue_access_token('DOMAIN_OWNER', self.user.owner_id, self.user.domain_id, **kwargs)
        refresh_token = self.issue_refresh_token('DOMAIN_OWNER', self.user.owner_id, self.user.domain_id, **kwargs)

        # Update user's last_accessed_at field
        self.user.update({'last_accessed_at': datetime.utcnow()})

        return {
            'access_token': access_token,
            'refresh_token': refresh_token
        }

    def refresh_token(self, user_id, domain_id, **kwargs):
        self.user = self.domain_owner_mgr.get_owner(owner_id=user_id, domain_id=domain_id)
        return self.issue_token(**kwargs)

    @staticmethod
    def _parse_password(credentials):
        pw_to_check = credentials.get('password', None)

        if pw_to_check is None:
            raise ERROR_INVALID_CREDENTIALS()

        return pw_to_check



PROTO = {
    'spaceone.identity.api.v1.domain': ['Domain'],
    'spaceone.identity.api.v1.domain_owner': ['DomainOwner'],
    'spaceone.identity.api.v1.endpoint': ['Endpoint'],
    'spaceone.identity.api.v1.provider': ['Provider'],
    'spaceone.identity.api.v1.service_account': ['ServiceAccount'],
    'spaceone.identity.api.v1.project_group': ['ProjectGroup'],
    'spaceone.identity.api.v1.project': ['Project'],
    'spaceone.identity.api.v1.policy': ['Policy'],
    'spaceone.identity.api.v1.role': ['Role'],
    'spaceone.identity.api.v1.role_binding': ['RoleBinding'],
    'spaceone.identity.api.v1.user': ['User'],
    'spaceone.identity.api.v1.api_key': ['APIKey'],
    'spaceone.identity.api.v1.token': ['Token'],
    'spaceone.identity.api.v1.authorization': ['Authorization']
}



import logging
from datetime import datetime

from spaceone.identity.error.error_authentication import *
from spaceone.identity.error.error_user import ERROR_USER_STATUS_CHECK_FAILURE
from spaceone.identity.lib.cipher import PasswordCipher
from spaceone.identity.manager.user_manager import UserManager
from spaceone.identity.manager.token_manager import JWTManager
from spaceone.identity.model import User

_LOGGER = logging.getLogger(__name__)


class LocalTokenManager(JWTManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user_mgr: UserManager = self.locator.get_manager('UserManager')

    def authenticate(self, user_id, domain_id, credentials):
        pw_to_check = self._parse_password(credentials)

        self.user = self.user_mgr.get_user(user_id, domain_id)

        self._check_user_state()

        # TODO: decrypt pw
        is_correct = PasswordCipher().checkpw(pw_to_check, self.user.password)
        _LOGGER.debug(f'[authenticate] is_correct: {is_correct}, pw_to_check: {pw_to_check}, hashed_pw: {self.user.password}')

        if is_correct:
            self.is_authenticated = True
        else:
            raise ERROR_AUTHENTICATION_FAILURE(user_id=self.user.user_id)

    def issue_temporary_token(self, user_id, domain_id, **kwargs):
        permissions = [
            'identity.User.get',
            'identity.User.update'
        ]

        # Issue token
        access_token = self.issue_access_token('USER', user_id, domain_id, permissions=permissions, **kwargs)

        return {'access_token': access_token}

    def issue_token(self, **kwargs):
        if self.is_authenticated is False:
            raise ERROR_NOT_AUTHENTICATED()

        permissions = self._get_permissions_from_required_actions()

        # Issue token
        access_token = self.issue_access_token('USER', self.user.user_id, self.user.domain_id,
                                               permissions=permissions, **kwargs)
        refresh_token = self.issue_refresh_token('USER', self.user.user_id, self.user.domain_id, **kwargs)

        # Update user's last_accessed_at field
        user = self.user.update({'last_accessed_at': datetime.utcnow()})

        return {
            'access_token': access_token,
            'refresh_token': refresh_token
        }

    def refresh_token(self, user_id, domain_id, **kwargs):
        self.user: User = self.user_mgr.get_user(user_id, domain_id)
        self._check_user_state()

        return self.issue_token(**kwargs)

    def _get_permissions_from_required_actions(self):
        if 'UPDATE_PASSWORD' in self.user.required_actions:
            return [
                'identity.User.get',
                'identity.User.update'
            ]

        return None

    @staticmethod
    def _parse_password(credentials):
        pw_to_check = credentials.get('password', None)

        if pw_to_check is None:
            raise ERROR_INVALID_CREDENTIALS()

        return pw_to_check

    def _check_user_state(self):
        if self.user.user_type == 'API_USER':
            raise ERROR_NOT_ALLOWED_ISSUE_TOKEN_API_USER(user_id=self.user.user_id)

        if self.user.state != 'ENABLED':
            raise ERROR_USER_STATUS_CHECK_FAILURE(user_id=self.user.user_id)

        if self.user.backend != 'LOCAL':
            raise ERROR_NOT_FOUND(key='user_id', value=self.user.user_id)



import logging

from google.protobuf.json_format import MessageToDict
from spaceone.core import pygrpc
from spaceone.core.connector import BaseConnector
from spaceone.core.utils import parse_grpc_endpoint

from spaceone.identity.error.error_authentication import *

_LOGGER = logging.getLogger(__name__)


class AuthPluginConnector(BaseConnector):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = None

    def initialize(self, endpoint):
        static_endpoint = self.config.get('endpoint')

        if static_endpoint:
            endpoint = static_endpoint

        _LOGGER.info(f'[initialize] endpoint: {endpoint}')

        e = parse_grpc_endpoint(endpoint)
        self.client = pygrpc.client(endpoint=e['endpoint'], ssl_enabled=e['ssl_enabled'])

    def call_login(self, credentials, options, secret_data, schema=None):
        params = {
            'options': options,
            'secret_data': secret_data,
            'schema': schema,
            'user_credentials': credentials
        }

        try:
            response = self.client.Auth.login(params, metadata=self.transaction.get_connection_meta())
        except ERROR_BASE as e:
            _LOGGER.error(f'[call_login] Auth.login failed. (reason={e.message})')
            raise ERROR_INVALID_CREDENTIALS()
        except Exception as e:
            _LOGGER.error(f'[call_login] Auth.login failed. (reason={str(e)})')
            raise ERROR_INVALID_CREDENTIALS()

        return MessageToDict(response, preserving_proto_field_name=True)

    def init(self, options):
        params = {
            'options': options
        }

        try:
            response = self.client.Auth.init(params, metadata=self.transaction.get_connection_meta())
            return MessageToDict(response, preserving_proto_field_name=True)
        except ERROR_BASE as e:
            raise ERROR_AUTHENTICATION_FAILURE_PLUGIN(message=e.message)
        except Exception as e:
            raise ERROR_AUTHENTICATION_FAILURE_PLUGIN(messsage=str(e))

    def verify(self, options, secret_data=None, schema=None):
        params = {
            'options': options,
            'secret_data': secret_data,
            'schema': schema
        }
        try:
            # TODO: meta (plugin has no meta)
            response = self.client.Auth.verify(params, metadata=self.transaction.get_connection_meta())
            return MessageToDict(response, preserving_proto_field_name=True)
        except ERROR_BASE as e:
            raise ERROR_AUTHENTICATION_FAILURE_PLUGIN(message=e.message)
        except Exception as e:
            raise ERROR_AUTHENTICATION_FAILURE_PLUGIN(messsage=str(e))

    def call_find(self, keyword, user_id, options, secret_data={}, schema=None):
        params = {
            'options': options,
            'secret_data': secret_data,
            'schema': schema,
            'keyword': keyword,
            'user_id': user_id
        }
        _LOGGER.info(f'[call_find] params: {params}')

        try:
            response = self.client.Auth.find(params, metadata=self.transaction.get_connection_meta())

            users_info = MessageToDict(response, preserving_proto_field_name=True)
            _LOGGER.debug(f'[call_find] MessageToDict(user_info): {users_info}')
            return users_info

        except ERROR_BASE as e:
            raise ERROR_AUTHENTICATION_FAILURE_PLUGIN(message=e.message)
        except Exception as e:
            raise ERROR_AUTHENTICATION_FAILURE_PLUGIN(messsage=str(e))



import functools
from spaceone.api.identity.v1 import project_group_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.identity.model.project_model import Project
from spaceone.identity.model.project_group_model import ProjectGroup
from spaceone.identity.info.role_info import RoleInfo

__all__ = ['ProjectGroupInfo', 'ProjectGroupsInfo', 'ProjectGroupRoleBindingInfo', 'ProjectGroupRoleBindingsInfo',
           'ProjectGroupProjectsInfo']


def ProjectGroupInfo(project_group_vo: ProjectGroup, minimal=False, parent_project_groups_info=None):
    info = {
        'project_group_id': project_group_vo.project_group_id,
        'name': project_group_vo.name
    }

    if not minimal:
        if parent_project_groups_info:
            parent_project_group_id = project_group_vo.parent_project_group_id
            if parent_project_group_id and parent_project_group_id in parent_project_groups_info:
                parent_project_group_info = ProjectGroupInfo(parent_project_groups_info[parent_project_group_id],
                                                             minimal=True)

                info.update({
                    'parent_project_group_info': parent_project_group_info
                })
        else:
            if project_group_vo.parent_project_group:
                info.update({
                    'parent_project_group_info': ProjectGroupInfo(project_group_vo.parent_project_group, minimal=True)
                })

        info.update({
            'tags': change_struct_type(project_group_vo.tags),
            'domain_id': project_group_vo.domain_id,
            'created_by': project_group_vo.created_by,
            'created_at': utils.datetime_to_iso8601(project_group_vo.created_at)
        })

    return project_group_pb2.ProjectGroupInfo(**info)


def ProjectGroupsInfo(project_group_vos, total_count, **kwargs):
    results = list(map(functools.partial(ProjectGroupInfo, **kwargs), project_group_vos))
    return project_group_pb2.ProjectGroupsInfo(results=results, total_count=total_count)


def ProjectGroupRoleBindingInfo(role_binding_vo, minimal=False):
    info = {
        'role_binding_id': role_binding_vo.role_binding_id,
        'resource_type': role_binding_vo.resource_type,
        'resource_id': role_binding_vo.resource_id,
        'role_info': RoleInfo(role_binding_vo.role, minimal=True) if role_binding_vo.role else None
    }

    if not minimal:
        info.update({
            'project_group_info': ProjectGroupInfo(role_binding_vo.project_group,
                                                   minimal=True) if role_binding_vo.project_group else None,
            'labels': change_list_value_type(role_binding_vo.labels),
            'tags': change_struct_type(role_binding_vo.tags),
            'domain_id': role_binding_vo.domain_id,
            'created_at': utils.datetime_to_iso8601(role_binding_vo.created_at)
        })

    return project_group_pb2.ProjectGroupRoleBindingInfo(**info)


def ProjectGroupRoleBindingsInfo(role_binding_vos, total_count, **kwargs):
    results = list(map(functools.partial(ProjectGroupRoleBindingInfo, **kwargs), role_binding_vos))

    return project_group_pb2.ProjectGroupRoleBindingsInfo(results=results, total_count=total_count)


def ProjectGroupProjectInfo(project_vo: Project, minimal=False, project_groups_info=None, **kwargs):
    info = {
        'project_id': project_vo.project_id,
        'name': project_vo.name
    }

    if not minimal:
        if project_groups_info:
            if project_vo.project_group_id and project_vo.project_group_id in project_groups_info:
                project_group_info = ProjectGroupInfo(project_groups_info[project_vo.project_group_id], minimal=True)
                info.update({
                    'project_group_info': project_group_info
                })
        else:
            if project_vo.project_group:
                info.update({
                    'project_group_info': ProjectGroupInfo(project_vo.project_group, minimal=True)
                })

        info.update({
            'tags': change_struct_type(project_vo.tags),
            'domain_id': project_vo.domain_id,
            'created_by': project_vo.created_by,
            'created_at': utils.datetime_to_iso8601(project_vo.created_at)
        })

    return project_group_pb2.ProjectGroupProjectInfo(**info)


def ProjectGroupProjectsInfo(project_vos, total_count, **kwargs):
    results = list(map(functools.partial(ProjectGroupProjectInfo, **kwargs), project_vos))
    return project_group_pb2.ProjectGroupProjectsInfo(results=results, total_count=total_count)



from spaceone.api.identity.v1 import token_pb2

__all__ = ['TokenInfo']


def TokenInfo(token_data):
    info = {
        'access_token': token_data['access_token'],
        'refresh_token': token_data['refresh_token']
    }

    return token_pb2.TokenInfo(**info)


import logging
from spaceone.core.service import *
from spaceone.core import utils
from spaceone.identity.error.error_project import *
from spaceone.identity.model.project_group_model import ProjectGroup
from spaceone.identity.manager.domain_manager import DomainManager
from spaceone.identity.manager.user_manager import UserManager
from spaceone.identity.manager.project_manager import ProjectManager
from spaceone.identity.manager.project_group_manager import ProjectGroupManager
from spaceone.identity.manager.role_manager import RoleManager
from spaceone.identity.manager.role_binding_manager import RoleBindingManager

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class ProjectService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.project_mgr: ProjectManager = self.locator.get_manager('ProjectManager')

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['name', 'project_group_id', 'domain_id'])
    def create(self, params):
        """ Create project

        Args:
            params (dict): {
                'name': 'str',
                'project_group_id': 'str',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            project_vo (object)
        """

        params['created_by'] = self.transaction.get_meta('user_id')

        if 'project_group_id' in params:
            project_group_mgr: ProjectGroupManager = self.locator.get_manager('ProjectGroupManager')
            params['project_group'] = project_group_mgr.get_project_group(params['project_group_id'],
                                                                          params['domain_id'])

        return self.project_mgr.create_project(params)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_id', 'domain_id'])
    def update(self, params):
        """ Update project

        Args:
            params (dict): {
                'project_id': 'str',
                'name': 'str',
                'project_group_id': 'str',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            project_vo (object)
        """

        domain_id = params['domain_id']

        project_vo = self.project_mgr.get_project(params['project_id'], domain_id)

        if 'project_group_id' in params:
            project_group_mgr: ProjectGroupManager = self.locator.get_manager('ProjectGroupManager')
            params['project_group'] = project_group_mgr.get_project_group(params['project_group_id'], domain_id)

        return self.project_mgr.update_project_by_vo(params, project_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_id', 'domain_id'])
    def delete(self, params):
        """ Delete project

        Args:
            params (dict): {
                'project_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        project_vo = self.project_mgr.get_project(params['project_id'], params['domain_id'])
        self.project_mgr.delete_project_by_vo(project_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_id', 'domain_id'])
    @change_only_key({'project_group_info': 'project_group'})
    def get(self, params):
        """ Get project

        Args:
            params (dict): {
                'project_id': 'str',
                'domain_id': 'str',
                'only': 'list'
            }

        Returns:
            project_vo (object)
        """

        return self.project_mgr.get_project(params['project_id'], params['domain_id'], params.get('only'))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    @change_only_key({'project_group_info': 'project_group'}, key_path='query.only')
    @append_query_filter(['project_id', 'name', 'project_group_id', 'domain_id', 'user_projects'])
    @append_keyword_filter(['project_id', 'name'])
    def list(self, params):
        """ List projects

        Args:
            params (dict): {
                'project_id': 'str',
                'name': 'str',
                'project_group_id': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)',
                'user_projects': 'list', # from meta
            }

        Returns:
            results (list): 'list of project_vo'
            total_count (int)
        """

        role_type = self.transaction.get_meta('authorization.role_type')
        user_projects = self.transaction.get_meta('authorization.projects')
        query = params.get('query', {})

        if 'only' in query and 'project_group' in query['only']:
            query['only'].append('project_group_id')

        if role_type == 'PROJECT':
            query['filter'].append({
                'k': 'user_projects',
                'v': user_projects,
                'o': 'in'
            })

        project_vos, total_count = self.project_mgr.list_projects(params.get('query', {}))

        return project_vos, total_count, self._get_project_groups_info(project_vos)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id', 'user_projects'])
    @append_keyword_filter(['project_id', 'name'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'user_projects': 'list', # from meta
            }

        Returns:
            values (list): 'list of statistics data'
            total_count (int)
        """

        query = params.get('query', {})
        return self.project_mgr.stat_projects(query)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_id', 'user_id', 'role_id', 'domain_id'])
    def add_member(self, params):
        """ Add project member

        Args:
            params (dict): {
                'project_id': 'str',
                'user_id': 'str',
                'is_external_user': 'bool',
                'role_id': 'str',
                'labels': 'list',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            role_binding_vo (object)
        """

        is_external_user = params.get('is_external_user', False)
        user_id = params['user_id']
        domain_id = params['domain_id']

        if is_external_user:
            self._create_external_user(user_id, domain_id)

        params['resource_type'] = 'identity.User'
        params['resource_id'] = user_id
        del params['user_id']

        role_mgr: RoleManager = self.locator.get_manager('RoleManager')
        role_binding_mgr: RoleBindingManager = self.locator.get_manager('RoleBindingManager')

        role_vo = role_mgr.get_role(params['role_id'], params['domain_id'])
        if role_vo.role_type != 'PROJECT':
            raise ERROR_ONLY_PROJECT_ROLE_TYPE_ALLOWED()

        return role_binding_mgr.create_role_binding(params)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_id', 'user_id', 'domain_id'])
    def modify_member(self, params):
        """ Modify project member

        Args:
            params (dict): {
                'project_id': 'str',
                'user_id': 'str',
                'labels': 'list',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            role_binding_vo (object)
        """

        project_id = params['project_id']
        user_id = params['user_id']
        domain_id = params['domain_id']

        project_vo = self.project_mgr.get_project(project_id, domain_id)

        role_binding_mgr: RoleBindingManager = self.locator.get_manager('RoleBindingManager')
        role_binding_vos = role_binding_mgr.get_project_role_binding('identity.User', user_id, domain_id,
                                                                     project_vo=project_vo)

        if role_binding_vos.count() == 0:
            raise ERROR_NOT_FOUND(key='user_id', value=user_id)

        return role_binding_mgr.update_role_binding_by_vo(params, role_binding_vos[0])

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_id', 'user_id', 'domain_id'])
    def remove_member(self, params):
        """ Remove project member

        Args:
            params (dict): {
                'project_id': 'str',
                'user_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        project_id = params['project_id']
        user_id = params['user_id']
        domain_id = params['domain_id']

        project_vo = self.project_mgr.get_project(project_id, domain_id)

        role_binding_mgr: RoleBindingManager = self.locator.get_manager('RoleBindingManager')
        role_binding_vos = role_binding_mgr.get_project_role_binding('identity.User', user_id, domain_id,
                                                                     project_vo=project_vo)

        if role_binding_vos.count() == 0:
            raise ERROR_NOT_FOUND(key='user_id', value=user_id)

        for role_binding_vo in role_binding_vos:
            role_binding_mgr.delete_role_binding_by_vo(role_binding_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['project_id', 'domain_id'])
    @change_only_key({'project_group_info': 'project_group', 'project_info': 'project', 'role_info': 'role'},
                     key_path='query.only')
    @append_query_filter(['user_id', 'role_id'])
    @append_keyword_filter(['user_id', 'name', 'email'])
    def list_members(self, params):
        """ List project members

        Args:
            params (dict): {
                'project_id': 'str',
                'user_id': 'str',
                'role_id': 'str',
                'domain_id': 'str',
                'include_parent_member': 'bool',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            results (list): 'list of role_binding_vo'
            total_count (int)
        """

        project_id = params['project_id']
        domain_id = params['domain_id']
        include_parent_member = params.get('include_parent_member', False)

        role_binding_mgr: RoleBindingManager = self.locator.get_manager('RoleBindingManager')

        query = params.get('query', {})
        query['filter'] = query.get('filter', [])

        if include_parent_member:
            project_vo = self.project_mgr.get_project(project_id, domain_id)
            project_group_vos = self._get_parents(project_vo.project_group, [])
            query['filter'].append({
                'k': 'project',
                'v': [project_vo, None],
                'o': 'in'
            })
            query['filter'].append({
                'k': 'project_group',
                'v': project_group_vos + [None],
                'o': 'in'
            })
            query['filter'].append({
                'k': 'role_type',
                'v': 'PROJECT',
                'o': 'eq'
            })
        else:
            query['filter'].append({
                'k': 'project_id',
                'v': project_id,
                'o': 'eq'
            })

        return role_binding_mgr.list_role_bindings(query)

    def _get_parents(self, project_group_vo: ProjectGroup, parent_project_group_vos):
        parent_project_group_vos.append(project_group_vo)

        if project_group_vo.parent_project_group:
            return self._get_parents(project_group_vo.parent_project_group, parent_project_group_vos)
        else:
            return parent_project_group_vos

    def _create_external_user(self, user_id, domain_id):
        user_mgr: UserManager = self.locator.get_manager('UserManager')
        domain_mgr: DomainManager = self.locator.get_manager('DomainManager')

        user_vos = user_mgr.filter_users(user_id=user_id, domain_id=domain_id)

        if user_vos.count() == 0:
            domain_vo = domain_mgr.get_domain(domain_id)
            user_mgr.create_user({
                'user_id': user_id,
                'user_type': 'USER',
                'backend': 'EXTERNAL',
                'domain_id': domain_id
            }, domain_vo)

    def _get_project_groups_info(self, project_vos):
        project_group_mgr: ProjectGroupManager = self.locator.get_manager('ProjectGroupManager')

        project_groups_info = {}
        project_group_ids = []
        for project_vo in project_vos:
            if project_vo.project_group_id:
                project_group_ids.append(project_vo.project_group_id)

        pg_vos = project_group_mgr.filter_project_groups(project_group_id=project_group_ids)
        for pg_vo in pg_vos:
            project_groups_info[pg_vo.project_group_id] = pg_vo

        return project_groups_info



import re
import logging

from spaceone.core import cache
from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector
from spaceone.identity.connector import AuthPluginConnector
from spaceone.identity.lib.cipher import PasswordCipher
from spaceone.identity.model import Domain
from spaceone.identity.model.user_model import User
from spaceone.identity.error.error_user import *
from spaceone.identity.manager.domain_manager import DomainManager

_LOGGER = logging.getLogger(__name__)


class UserManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user_model: User = self.locator.get_model('User')

    def create_user(self, params, domain_vo: Domain, is_first_login_user=False):
        def _rollback(user_vo):
            _LOGGER.info(f'[create_user._rollback] Delete user : {user_vo.name} ({user_vo.user_id})')
            user_vo.delete()

        params['state'] = params.get('state', 'ENABLED')

        # If user create external authentication, call find action.
        if params['backend'] == 'EXTERNAL':
            if not is_first_login_user:
                found_users, count = self.find_user(
                    {
                        'user_id': params['user_id']
                    },
                    domain_vo
                )

                if count == 1:
                    found_user = found_users[0]
                    _LOGGER.debug(f'[create_user] found user: {found_user}')

                    if found_user.get('state') in ['ENABLED', 'DISABLED']:
                        params['state'] = found_user['state']
                    else:
                        params['state'] = 'PENDING'

                    if 'name' not in params:
                        params['name'] = found_user.get('name')

                    if 'email' not in params:
                        params['email'] = found_user.get('email')

                elif count > 1:
                    raise ERROR_TOO_MANY_USERS_IN_EXTERNAL_AUTH(user_id=params['user_id'])
                else:
                    raise ERROR_NOT_FOUND_USER_IN_EXTERNAL_AUTH(user_id=params['user_id'])

        else:
            if params['user_type'] == 'API_USER':
                params['password'] = None
            else:
                self._check_user_id_format(params['user_id'])

                password = params.get('password')
                if password:
                    self._check_password_format(password)
                else:
                    raise ERROR_REQUIRED_PARAMETER(key='password')

                hashed_pw = PasswordCipher().hashpw(password)
                params['password'] = hashed_pw

        user_name = params.get('name')
        user_email = params.get('email')

        if user_name:
            params['name'] = user_name.strip()
        else:
            params['name'] = ''

        if user_email:
            params['email'] = user_email.strip()
        else:
            params['email'] = ''

        user_vo = self.user_model.create(params)

        self.transaction.add_rollback(_rollback, user_vo)

        return user_vo

    def update_user(self, params):
        user_vo: User = self.get_user(params['user_id'], params['domain_id'])
        return self.update_user_by_vo(params, user_vo)

    def update_user_by_vo(self, params, user_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[update_user._rollback] Revert Data : {old_data["name"], ({old_data["user_id"]})}')
            user_vo.update(old_data)

        required_actions = list(user_vo.required_actions)
        is_change_required_actions = False

        if new_password := params.get('password'):
            if PasswordCipher().checkpw(new_password, user_vo.password):
                raise ERROR_PASSWORD_NOT_CHANGED(user_id=user_vo.user_id)

            self._check_password_format(params['password'])
            hashed_pw = PasswordCipher().hashpw(params['password'])
            params['password'] = hashed_pw

            if 'UPDATE_PASSWORD' in required_actions:
                required_actions.remove('UPDATE_PASSWORD')
                is_change_required_actions = True

        if is_change_required_actions:
            params['required_actions'] = required_actions

        self.transaction.add_rollback(_rollback, user_vo.to_dict())

        return user_vo.update(params)

    def delete_user(self, user_id, domain_id):
        user_vo = self.get_user(user_id, domain_id)
        user_vo.delete()

        cache.delete_pattern(f'user-state:{domain_id}:{user_id}')
        cache.delete_pattern(f'role-bindings:{domain_id}:{user_id}*')
        cache.delete_pattern(f'user-permissions:{domain_id}:{user_id}*')
        cache.delete_pattern(f'user-scopes:{domain_id}:{user_id}*')

    def enable_user(self, user_id, domain_id):
        def _rollback(old_data):
            _LOGGER.info(f'[enable_user._rollback] Revert Data : {old_data}')
            user_vo.update(old_data)

        user_vo: User = self.get_user(user_id, domain_id)

        if user_vo.state != 'ENABLED':
            self.transaction.add_rollback(_rollback, user_vo.to_dict())
            user_vo.update({'state': 'ENABLED'})

            cache.delete_pattern(f'user-state:{domain_id}:{user_id}')

        return user_vo

    def disable_user(self, user_id, domain_id):
        def _rollback(old_data):
            _LOGGER.info(f'[disable_user._rollback] Revert Data : {old_data}')
            user_vo.update(old_data)

        user_vo: User = self.get_user(user_id, domain_id)

        if user_vo.state != 'DISABLED':
            self.transaction.add_rollback(_rollback, user_vo.to_dict())
            user_vo.update({'state': 'DISABLED'})

            cache.delete_pattern(f'user-state:{domain_id}:{user_id}')

        return user_vo

    def get_user(self, user_id, domain_id, only=None):
        return self.user_model.get(user_id=user_id, domain_id=domain_id, only=only)

    def filter_users(self, **conditions):
        return self.user_model.filter(**conditions)

    def list_users(self, query):
        return self.user_model.query(**query)

    def stat_users(self, query):
        return self.user_model.stat(**query)

    def find_user(self, search, domain_vo: Domain):
        keyword = search.get('keyword')
        user_id = search.get('user_id')

        domain_mgr: DomainManager = self.locator.get_manager('DomainManager')

        endpoint = domain_mgr.get_auth_plugin_endpoint_by_vo(domain_vo)

        response = self._call_find(keyword, user_id, domain_vo, endpoint)
        results = response.get('results', [])
        total_count = response.get('total_count', 0)

        return results, total_count

    @staticmethod
    def _check_user_id_format(user_id):
        rule = r"[^@]+@[^@]+\.[^@]+"
        if not re.match(rule, user_id):
            raise ERROR_INCORRECT_USER_ID_FORMAT(rule='Email format required.')

    @staticmethod
    def _check_password_format(password):
        if len(password) < 8:
            raise ERROR_INCORRECT_PASSWORD_FORMAT(rule='At least 9 characters long.')
        elif not re.search("[a-z]", password):
            raise ERROR_INCORRECT_PASSWORD_FORMAT(rule='Contains at least one lowercase character')
        elif not re.search("[A-Z]", password):
            raise ERROR_INCORRECT_PASSWORD_FORMAT(rule='Contains at least one uppercase character')
        elif not re.search("[0-9]", password):
            raise ERROR_INCORRECT_PASSWORD_FORMAT(rule='Contains at least one number')

    def _call_find(self, keyword, user_id, domain_vo, endpoint):
        options = domain_vo.plugin_info.options

        auth_plugin_conn: AuthPluginConnector = self.locator.get_connector('AuthPluginConnector')
        auth_plugin_conn.initialize(endpoint)

        secret_data, schema = self._get_auth_plugin_secret(domain_vo.to_dict())
        return auth_plugin_conn.call_find(keyword, user_id, options, secret_data, schema)

    def _get_auth_plugin_secret(self, domain_data):
        """
        Return: (secret_data, schema)
                Default: ({}, None)
        """
        domain_id = domain_data['domain_id']
        plugin_info = domain_data.get('plugin_info', {})
        secret_id = plugin_info.get('secret_id')

        if secret_id is None:
            return {}, None

        secret_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', service='secret')
        secret = secret_connector.dispatch('Secret.get', {'secret_id': secret_id, 'domain_id': domain_id})
        secret_data = secret_connector.dispatch('Secret.get_data', {'secret_id': secret_id, 'domain_id': domain_id})

        return secret_data.get('data', {}), secret_data.get('schema')



import logging
from spaceone.core.auth.jwt import JWTUtil
from spaceone.core.cache import cacheable
from spaceone.core.manager import *
from spaceone.core import utils
from spaceone.identity.model.domain_secret_model import DomainSecret
from spaceone.identity.model.domain_model import Domain

_LOGGER = logging.getLogger(__name__)


class DomainSecretManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.domain_secret_model: DomainSecret = self.locator.get_model('DomainSecret')

    def create_domain_secret(self, domain_id):
        def _rollback(vo):
            _LOGGER.info(f'[create_domain_secret._rollback] Delete domain-secret : {vo.domain_id} ({vo.secret[0:6]}...)')
            vo.delete()

        # Generate Domain-secret
        secret = self._generate_domain_secret(domain_id)

        # Generate Domain Key
        secret['domain_key'] = utils.random_string(16)

        domain_secret_vo: DomainSecret = self.domain_secret_model.create(secret)
        self.transaction.add_rollback(_rollback, domain_secret_vo)

    @cacheable(key='public-jwk:{domain_id}', expire=600)
    def get_domain_public_key(self, domain_id):
        domain_secret_vo: DomainSecret = self.domain_secret_model.get(domain_id=domain_id)
        return domain_secret_vo.pub_jwk

    @cacheable(key='private-jwk:{domain_id}', expire=600)
    def get_domain_private_key(self, domain_id):
        domain_secret_vo: DomainSecret = self.domain_secret_model.get(domain_id=domain_id)
        return domain_secret_vo.prv_jwk

    @cacheable(key='refresh-public-jwk:{domain_id}', expire=600)
    def get_domain_refresh_public_key(self, domain_id):
        domain_secret_vo: DomainSecret = self.domain_secret_model.get(domain_id=domain_id)
        if not domain_secret_vo.refresh_pub_jwk:
            domain_secret_vo: DomainSecret = self._create_refresh_key(domain_secret_vo, domain_id)

        return domain_secret_vo.refresh_pub_jwk

    @cacheable(key='refresh-private-jwk:{domain_id}', expire=600)
    def get_domain_refresh_private_key(self, domain_id):
        domain_secret_vo: DomainSecret = self.domain_secret_model.get(domain_id=domain_id)
        if not domain_secret_vo.refresh_pub_jwk:
            domain_secret_vo: DomainSecret = self._create_refresh_key(domain_secret_vo, domain_id)

        return domain_secret_vo.refresh_prv_jwk

    def _create_refresh_key(self, domain_secret_vo: DomainSecret, domain_id):
        domain_model: Domain = self.locator.get_model('Domain')
        domain_vo = domain_model.get(domain_id=domain_id)
        refresh_private_jwk, refresh_public_jwk = JWTUtil.generate_jwk()

        return domain_secret_vo.update({
            'refresh_pub_jwk': refresh_public_jwk,
            'refresh_prv_jwk': refresh_private_jwk,
            'domain': domain_vo,
        })

    def _generate_domain_secret(self, domain_id: str) -> dict:
        domain_model: Domain = self.locator.get_model('Domain')
        domain_vo = domain_model.get(domain_id=domain_id)

        private_jwk, public_jwk = JWTUtil.generate_jwk()
        refresh_private_jwk, refresh_public_jwk = JWTUtil.generate_jwk()
        data = {
            'pub_jwk': public_jwk,
            'prv_jwk': private_jwk,
            'refresh_pub_jwk': refresh_public_jwk,
            'refresh_prv_jwk': refresh_private_jwk,
            'domain_id': domain_id,
            'domain': domain_vo
        }
        return data



import functools
from spaceone.api.identity.v1 import service_account_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.identity.model.service_account_model import ServiceAccount
from spaceone.identity.info.project_info import ProjectInfo

__all__ = ['ServiceAccountInfo', 'ServiceAccountsInfo']


def ServiceAccountInfo(service_account_vo: ServiceAccount, minimal=False, projects_info=None):
    info = {
        'service_account_id': service_account_vo.service_account_id,
        'name': service_account_vo.name,
        'provider': service_account_vo.provider,
        'service_account_type': service_account_vo.service_account_type
    }

    if not minimal:
        info.update({
            'trusted_service_account_id': service_account_vo.trusted_service_account_id,
            'scope': service_account_vo.scope,
            'data': change_struct_type(service_account_vo.data),
            'tags': change_struct_type(service_account_vo.tags),
            'domain_id': service_account_vo.domain_id,
            'created_at': utils.datetime_to_iso8601(service_account_vo.created_at)
        })

        if projects_info:
            if service_account_vo.project_id and service_account_vo.project_id in projects_info:
                project_info = ProjectInfo(projects_info[service_account_vo.project_id], minimal=True)
                info.update({
                    'project_info': project_info
                })
        else:
            if service_account_vo.project:
                info.update({
                    'project_info': ProjectInfo(service_account_vo.project, minimal=True)
                })

    return service_account_pb2.ServiceAccountInfo(**info)


def ServiceAccountsInfo(service_account_vos, total_count, **kwargs):
    results = list(map(functools.partial(ServiceAccountInfo, **kwargs), service_account_vos))

    return service_account_pb2.ServiceAccountsInfo(results=results, total_count=total_count)



import logging
from spaceone.core import cache
from spaceone.core.manager import BaseManager
from spaceone.identity.manager import ProjectGroupManager
from spaceone.identity.model.project_model import Project

_LOGGER = logging.getLogger(__name__)


class ProjectManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.project_model: Project = self.locator.get_model('Project')
        self.project_group_mgr: ProjectGroupManager = self.locator.get_manager('ProjectGroupManager')

    def create_project(self, params):
        def _rollback(project_vo):
            _LOGGER.info(f'[create_project._rollback] Delete project : {project_vo.name} ({project_vo.project_id})')
            project_vo.delete()

        project_vo: Project = self.project_model.create(params)
        self.transaction.add_rollback(_rollback, project_vo)

        project_group_id = params['project_group_id']
        cache.delete_pattern(f'project-path:*{project_group_id}*')
        cache.delete_pattern(f'role-bindings:*{project_group_id}*')
        cache.delete_pattern(f'user-scopes:*{project_group_id}*')
        self._delete_parent_project_group_cache(params['project_group'])

        return project_vo

    def update_project(self, params):
        project_vo = self.get_project(params['project_id'], params['domain_id'])
        return self.update_project_by_vo(params, project_vo)

    def update_project_by_vo(self, params, project_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[update_project._rollback] Revert Data : {old_data["name"]} ({old_data["project_id"]})')
            project_vo.update(old_data)

        self.transaction.add_rollback(_rollback, project_vo.to_dict())

        if 'project_group' in params:
            new_project_group_id = params['project_group_id']
            old_project_group_id = project_vo.project_group.project_group_id

            cache.delete_pattern(f'project-path:*{new_project_group_id}*')
            cache.delete_pattern(f'project-path:*{old_project_group_id}*')
            cache.delete_pattern(f'role-bindings:*{new_project_group_id}*')
            cache.delete_pattern(f'role-bindings:*{old_project_group_id}*')
            cache.delete_pattern(f'user-scopes:*{new_project_group_id}*')
            cache.delete_pattern(f'user-scopes:*{old_project_group_id}*')
            self._delete_parent_project_group_cache(params['project_group'])
            self._delete_parent_project_group_cache(project_vo.project_group)

        return project_vo.update(params)

    def delete_project(self, project_id, domain_id):
        project_vo = self.get_project(project_id, domain_id)
        self.delete_project_by_vo(project_vo)

    def delete_project_by_vo(self, project_vo):
        project_group_id = project_vo.project_group_id
        project_vo.delete()

        if project_group_id:
            cache.delete_pattern(f'project-path:*{project_group_id}*')
            cache.delete_pattern(f'project-group-children:*{project_group_id}')
            cache.delete_pattern(f'role-bindings:*{project_group_id}*')
            cache.delete_pattern(f'role-bindings:*:')
            cache.delete_pattern(f'user-scopes:*{project_group_id}*')
            self._delete_parent_project_group_cache(project_vo.project_group)

    def get_project(self, project_id, domain_id, only=None):
        return self.project_model.get(project_id=project_id, domain_id=domain_id, only=only)

    def filter_projects(self, **conditions):
        return self.project_model.filter(**conditions)

    def list_projects(self, query):
        return self.project_model.query(**query)

    def stat_projects(self, query):
        return self.project_model.stat(**query)

    @cache.cacheable(key='project-path:{domain_id}:{project_id}:{project_group_id}', expire=3600)
    def get_project_path(self, project_id, project_group_id, domain_id):
        project_path = None
        if project_id:
            try:
                project_vo = self.get_project(project_id, domain_id)
                project_path = [project_id]
                project_path += self._get_parent_project_group_path(project_vo.project_group, [])
            except Exception as e:
                _LOGGER.debug(f'[get_project_path] Project could not be found. '
                              f'(project_id={project_id}, reason={e})')

        elif project_group_id:
            try:
                project_group_vo = self.project_group_mgr.get_project_group(project_group_id, domain_id)
                project_path = [project_group_id]
                project_path += self._get_parent_project_group_path(project_group_vo.parent_project_group, [])
            except Exception as e:
                _LOGGER.debug(f'[_get_project_path] Project group could not be found. '
                              f'(project_group_id={project_group_id}, reason={e})')

        return project_path

    def _get_parent_project_group_path(self, project_group_vo, project_path):
        project_group_id = project_group_vo.project_group_id
        project_path.append(project_group_id)

        if project_group_vo.parent_project_group:
            project_path = self._get_parent_project_group_path(project_group_vo.parent_project_group, project_path)

        return project_path

    def _delete_parent_project_group_cache(self, project_group_vo):
        cache.delete_pattern(f'project-group-children:*{project_group_vo.project_group_id}')
        if project_group_vo.parent_project_group:
            self._delete_parent_project_group_cache(project_group_vo.parent_project_group)



from mongoengine import *
from spaceone.core.model.mongo_model import MongoModel


class Provider(MongoModel):
    provider = StringField(max_length=40, unique=True)
    name = StringField(max_length=255)
    template = DictField()
    metadata = DictField()
    capability = DictField()
    tags = DictField()
    created_at = DateTimeField(auto_now_add=True)

    meta = {
        'updatable_fields': [
            'name',
            'template',
            'metadata',
            'capability',
            'tags'
        ],
        'minimal_fields': [
            'provider',
            'name'
        ],
        'ordering': ['created_at'],
        'indexes': [
            # 'provider',
        ]
    }



from spaceone.core.service import *
from spaceone.core.error import *
from spaceone.core import config
from spaceone.identity.manager import RoleManager, PolicyManager, DomainManager


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class RoleService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.role_mgr: RoleManager = self.locator.get_manager('RoleManager')

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['name', 'role_type', 'policies', 'domain_id'])
    def create(self, params):
        """ Create role

        Args:
            params (dict): {
                'name': 'str',
                'role_type': 'str',
                'policies': 'list',
                'page_permissions': 'list',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            role_vo (object)
        """

        # The system role type is allowed only in the root domain.
        if params['role_type'] == 'SYSTEM':
            self._check_system_role_type(params['domain_id'])

        params['policies'] = self._check_policy_info(params['policies'], params['domain_id'])

        return self.role_mgr.create_role(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['role_id', 'domain_id'])
    def update(self, params):
        """ Update role

        Args:
            params (dict): {
                'role_id': 'str',
                'name': 'str',
                'policies': 'list',
                'page_permissions': 'list',
                'tags': 'dict',
                'release_page_permissions': 'bool',
                'domain_id': 'str'
            }

        Returns:
            role_vo (object)
        """

        release_page_permissions = params.get('release_page_permissions', False)

        if release_page_permissions:
            params['page_permissions'] = []

        if 'policies' in params:
            params['policies'] = self._check_policy_info(params['policies'], params['domain_id'])

        return self.role_mgr.update_role(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['role_id', 'domain_id'])
    def delete(self, params):
        """ Delete role

        Args:
            params (dict): {
                'role_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        self.role_mgr.delete_role(params['role_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['role_id', 'domain_id'])
    def get(self, params):
        """ Get role

        Args:
            params (dict): {
                'role_id': 'str',
                'domain_id': 'str',
                'only': 'list'
            }

        Returns:
            role_vo (object)
        """

        return self.role_mgr.get_role(params['role_id'], params['domain_id'], params.get('only'))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    @append_query_filter(['role_id', 'name', 'role_type', 'policy_id', 'domain_id'])
    @append_keyword_filter(['role_id', 'name'])
    def list(self, params):
        """ List roles

        Args:
            params (dict): {
                'role_id': 'str',
                'name': 'str',
                'role_type': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            results (list): 'list of role_vo'
            total_count (int)
        """

        return self.role_mgr.list_roles(params.get('query', {}))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id'])
    @append_keyword_filter(['role_id', 'name'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
            }

        Returns:
            values (list): 'list of statistics data'
            total_count (int)
        """

        query = params.get('query', {})
        return self.role_mgr.stat_roles(query)

    def _check_policy_info(self, policies, domain_id):
        policy_mgr: PolicyManager = self.locator.get_manager('PolicyManager')

        change_policies = []
        for policy in policies:
            if policy['policy_type'] == 'MANAGED':
                policy['policy'] = policy_mgr.get_managed_policy(policy['policy_id'], domain_id)
            elif policy['policy_type'] == 'CUSTOM':
                policy['policy'] = policy_mgr.get_policy(policy['policy_id'], domain_id)
            change_policies.append(policy)

        return change_policies

    def _check_system_role_type(self, domain_id):
        root_domain_name = config.get_global('ROOT_DOMAIN_NAME', 'root')
        domain_mgr: DomainManager = self.locator.get_manager('DomainManager')
        domain_vo = domain_mgr.get_domain(domain_id)

        if root_domain_name != domain_vo.name:
            raise ERROR_PERMISSION_DENIED()



import logging

from spaceone.core import cache
from spaceone.core.service import *
from spaceone.core.error import *
from spaceone.identity.manager.authorization_manager import AuthorizationManager
from spaceone.identity.manager.domain_manager import DomainManager
from spaceone.identity.manager.user_manager import UserManager
from spaceone.identity.manager.project_manager import ProjectManager
from spaceone.identity.manager.project_group_manager import ProjectGroupManager
from spaceone.identity.manager.role_binding_manager import RoleBindingManager
from spaceone.identity.manager.role_manager import RoleManager
from spaceone.identity.manager.policy_manager import PolicyManager
from spaceone.identity.conf.permissions_conf import DEFAULT_PERMISSIONS

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@event_handler
class AuthorizationService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.auth_mgr: AuthorizationManager = self.locator.get_manager('AuthorizationManager')
        self.project_mgr: ProjectManager = self.locator.get_manager('ProjectManager')
        self.project_group_mgr: ProjectGroupManager = self.locator.get_manager('ProjectGroupManager')

    @transaction(append_meta={'authorization.scope': 'SYSTEM'})
    @check_required(['service', 'resource', 'verb', 'scope'])
    def verify(self, params):
        """ Verify authority

        Args:
            params (dict): {
                'service': 'str',
                'resource': 'str',
                'verb': 'str',
                'scope': 'str',
                'domain_id': 'str',
                'project_id': 'str',
                'project_group_id': 'str',
                'user_id': 'str',
                'require_project_id': 'bool',
                'require_project_group_id': 'bool',
                'require_user_id': 'bool',
                'require_domain_id': 'bool',
            }

        Returns:
            domain_owner_vo (object)
        """

        user_id = self.transaction.get_meta('user_id')
        domain_id = self.transaction.get_meta('domain_id')
        service = params['service']
        resource = params['resource']
        verb = params['verb']
        scope = params['scope']
        request_domain_id = params.get('domain_id')
        request_project_id = params.get('project_id')
        request_project_group_id = params.get('project_group_id')
        request_user_id = params.get('user_id')
        require_project_id = params.get('require_project_id', False)
        require_project_group_id = params.get('require_project_group_id', False)
        require_user_id = params.get('require_user_id', False)
        require_domain_id = params.get('require_domain_id', False)

        self._check_user_state(user_id, domain_id)
        self._check_domain_state(domain_id)

        project_path = self.project_mgr.get_project_path(request_project_id, request_project_group_id, domain_id)
        role_type, request_roles, projects, project_groups = \
            self._get_user_roles_and_projects_from_role_bindings(user_id, domain_id, scope, project_path)

        projects, project_groups = self._get_all_projects_and_project_groups(projects, project_groups, domain_id)

        user_permissions = self._get_permissions(role_type, request_roles, domain_id)

        self.auth_mgr.check_permissions(user_id, domain_id, user_permissions, service, resource, verb, request_roles)
        self.auth_mgr.check_scope_by_role_type(user_id, domain_id, scope, role_type, projects, project_groups,
                                               request_domain_id, request_project_id, request_project_group_id,
                                               request_user_id, require_project_id, require_project_group_id,
                                               require_user_id, require_domain_id)

        return {
            'role_type': role_type,
            'projects': projects,
            'project_groups': project_groups
        }

    @cache.cacheable(key='user-state:{domain_id}:{user_id}', expire=3600)
    def _check_user_state(self, user_id, domain_id):
        user_mgr: UserManager = self.locator.get_manager('UserManager')
        try:
            user_vo = user_mgr.get_user(user_id, domain_id)
        except Exception as e:
            _LOGGER.error(f'[_check_user_state] Get User Error: {e}')
            raise ERROR_PERMISSION_DENIED()

        if user_vo.state == 'DISABLED':
            _LOGGER.error(f'[_check_user_state] User has been disabled. (user_id={user_id}, domain_id={domain_id})')
            raise ERROR_PERMISSION_DENIED()

    @cache.cacheable(key='domain-state:{domain_id}', expire=3600)
    def _check_domain_state(self, domain_id):
        domain_mgr: DomainManager = self.locator.get_manager('DomainManager')
        try:
            domain_vo = domain_mgr.get_domain(domain_id)
        except Exception as e:
            _LOGGER.error(f'[_check_user_state] Get Domain Error: {e}')
            raise ERROR_PERMISSION_DENIED()

        if domain_vo.state == 'DISABLED':
            _LOGGER.error(f'[_check_domain_state] Domain has been disabled. (domain_id={domain_id})')
            raise ERROR_PERMISSION_DENIED()

    @cache.cacheable(key='role-bindings:{domain_id}:{user_id}:{scope}:{project_path}', expire=3600)
    def _get_user_roles_and_projects_from_role_bindings(self, user_id, domain_id, scope, project_path):
        role_binding_mgr: RoleBindingManager = self.locator.get_manager('RoleBindingManager')

        role_binding_vos = role_binding_mgr.get_user_role_bindings(user_id, domain_id)

        role_type = None
        request_roles = []
        projects = []
        project_groups = []
        for role_binding_vo in role_binding_vos:
            rb_role_type = role_binding_vo.role.role_type
            rb_role_id = role_binding_vo.role_id
            if role_type is None:
                role_type = rb_role_type
            else:
                if rb_role_type != 'PROJECT':
                    role_type = rb_role_type

            if rb_role_type == 'PROJECT':
                if role_binding_vo.project_id:
                    project_id = role_binding_vo.project_id
                    projects.append(project_id)

                    if scope == 'PROJECT' and isinstance(project_path, list):
                        if project_id in project_path:
                            request_roles.append(rb_role_id)
                    else:
                        request_roles.append(rb_role_id)

                elif role_binding_vo.project_group_id:
                    project_group_id = role_binding_vo.project_group_id
                    project_groups.append(project_group_id)

                    if scope == 'PROJECT' and isinstance(project_path, list):
                        if project_group_id in project_path:
                            request_roles.append(rb_role_id)
                    else:
                        request_roles.append(rb_role_id)
            else:
                request_roles.append(rb_role_id)

        return role_type or 'USER', list(set(request_roles)), list(set(projects)), list(set(project_groups))

    def _get_request_roles_by_scope(self, user_role_type, user_roles, scope, project_path):
        if user_role_type == 'SYSTEM':
            return user_roles['SYSTEM']
        elif user_role_type == 'DOMAIN':
            if scope == 'PROJECT':
                for project_or_project_group_id in project_path:
                    if project_or_project_group_id in user_roles['PROJECT']:
                        return user_roles['PROJECT'][project_or_project_group_id]

            return user_roles['DOMAIN']

        elif user_role_type == 'PROJECT':
            if scope == 'PROJECT':
                if len(project_path) > 0:
                    for project_or_project_group_id in project_path:
                        if project_or_project_group_id in user_roles['PROJECT']:
                            return user_roles['PROJECT'][project_or_project_group_id]

                    raise ERROR_PERMISSION_DENIED()

            project_roles = []
            for project_or_project_group_id, roles in user_roles['PROJECT'].items():
                project_roles += roles

            return project_roles
        else:
            return []

    def _get_all_projects_and_project_groups(self, projects, project_groups, domain_id):
        for project_group_id in project_groups[:]:
            child_projects, child_project_groups = self._get_children_in_project_group(project_group_id, domain_id)
            projects += child_projects
            project_groups += child_project_groups

        return projects, project_groups

    @cache.cacheable(key='project-group-children:{domain_id}:{project_group_id}', expire=3600)
    def _get_children_in_project_group(self, project_group_id, domain_id):
        project_group_vo = self.project_group_mgr.get_project_group(project_group_id, domain_id)
        related_project_groups = self._get_related_project_group(project_group_vo, [project_group_vo])

        project_vos = self.project_mgr.filter_projects(project_group=related_project_groups)
        project_group_vos = self.project_group_mgr.filter_project_groups(parent_project_group=related_project_groups)

        projects = [project_vo.project_id for project_vo in project_vos]
        project_groups = [project_group_vo.project_group_id for project_group_vo in project_group_vos]

        return projects, project_groups

    def _get_related_project_group(self, project_group_vo, related_project_groups):
        project_group_vos = self.project_group_mgr.filter_project_groups(parent_project_group=project_group_vo)

        if project_group_vos.count() > 0:
            related_project_groups += project_group_vos
            for pg_vo in project_group_vos:
                related_project_groups = self._get_related_project_group(pg_vo, related_project_groups)

        return related_project_groups

    @cache.cacheable(key='role-permissions:{domain_id}:{role_id}', expire=3600)
    def _get_role_permissions(self, role_id, domain_id):
        role_mgr: RoleManager = self.locator.get_manager('RoleManager')
        policy_mgr: PolicyManager = self.locator.get_manager('PolicyManager')
        role_vo = role_mgr.get_role(role_id, domain_id)
        permissions = []

        for role_policy in role_vo.policies:
            if role_policy.policy_type == 'MANAGED':
                policy_vo = policy_mgr.get_managed_policy(role_policy.policy.policy_id, domain_id)
                permissions += policy_vo.permissions

            elif role_policy.policy_type == 'CUSTOM':
                permissions += role_policy.policy.permissions

        return permissions

    def _get_permissions(self, user_role_type, roles, domain_id):
        if user_role_type == 'USER':
            return DEFAULT_PERMISSIONS
        else:
            permissions = []

            for role_id in roles:
                permissions += self._get_role_permissions(role_id, domain_id)

            return permissions



import logging
from datetime import datetime

from spaceone.core import cache
from spaceone.core.manager import BaseManager
from spaceone.core import utils
from spaceone.core.connector.space_connector import SpaceConnector
from spaceone.identity.error.error_role import *
from spaceone.identity.model.policy_model import Policy
from spaceone.identity.manager.role_manager import RoleManager

_LOGGER = logging.getLogger(__name__)


class PolicyManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.policy_model: Policy = self.locator.get_model('Policy')

    def create_policy(self, params):
        def _rollback(policy_vo):
            _LOGGER.info(f'[create_policy._rollback] Create policy : {policy_vo.name} ({policy_vo.policy_id})')
            policy_vo.delete()

        policy_vo: Policy = self.policy_model.create(params)
        self.transaction.add_rollback(_rollback, policy_vo)

        return policy_vo

    def update_policy(self, params):
        def _rollback(old_data):
            _LOGGER.info(f'[update_policy._rollback] Revert Data : {old_data["name"]} ({old_data["policy_id"]})')
            policy_vo.update(old_data)

        policy_vo: Policy = self.get_policy(params['policy_id'], params['domain_id'])
        self.transaction.add_rollback(_rollback, policy_vo.to_dict())

        params['updated_at'] = datetime.utcnow()
        policy_vo = policy_vo.update(params)

        if 'permissions' in params:
            self._delete_role_cache(policy_vo)

        return policy_vo

    def delete_policy(self, policy_id, domain_id):
        policy_vo: Policy = self.get_policy(policy_id, domain_id)
        role_vos = self._get_roles_using_policy(policy_vo)

        for role_vo in role_vos:
            raise ERROR_POLICY_IS_IN_USE(policy_id=policy_id, role_id=role_vo.role_id)

        policy_vo.delete()

    def get_policy(self, policy_id, domain_id, only=None):
        return self.policy_model.get(policy_id=policy_id, domain_id=domain_id, policy_type='CUSTOM', only=only)

    def list_policies(self, query):
        return self.policy_model.query(**query)

    def stat_policies(self, query):
        return self.policy_model.stat(**query)

    def get_managed_policy(self, policy_id, domain_id):
        repo_managed_policy_info = self._get_managed_policy_from_repository(policy_id, domain_id)
        local_managed_policy_vo = self._get_managed_policy_from_local(policy_id, domain_id)

        if repo_managed_policy_info:
            if local_managed_policy_vo:
                if repo_managed_policy_info.get('updated_at') == local_managed_policy_vo.updated_at:
                    return local_managed_policy_vo

                return self._update_managed_policy(local_managed_policy_vo, repo_managed_policy_info)

            return self._create_managed_policy(policy_id, domain_id, repo_managed_policy_info)
        else:
            if local_managed_policy_vo:
                return local_managed_policy_vo

            raise ERROR_NOT_FOUND(key='policy_id', value=policy_id)

    def _create_managed_policy(self, policy_id, domain_id, repo_managed_policy_info):
        policy_vo: Policy = self.create_policy({
            'policy_id': policy_id,
            'name': repo_managed_policy_info['name'],
            'policy_type': 'MANAGED',
            'permissions': repo_managed_policy_info.get('permissions', []),
            'tags': repo_managed_policy_info.get('tags', []),
            'domain_id': domain_id,
            'updated_at': repo_managed_policy_info.get('updated_at')
        })

        return policy_vo

    def _update_managed_policy(self, local_managed_policy_vo, repo_managed_policy_info):
        policy_vo: Policy = local_managed_policy_vo.update({
            'name': repo_managed_policy_info['name'],
            'permissions': repo_managed_policy_info.get('permissions', []),
            'tags': repo_managed_policy_info.get('tags', []),
            'updated_at': repo_managed_policy_info.get('updated_at')
        })

        self._delete_role_cache(policy_vo)

        return policy_vo

    @cache.cacheable(key='managed-policy:{domain_id}:{policy_id}', expire=600)
    def _get_managed_policy_from_repository(self, policy_id, domain_id):
        repo_connector: SpaceConnector = self.locator.get_connector('SpaceConnector', service='repository')
        try:
            return repo_connector.dispatch('Policy.get', {'policy_id': policy_id, 'domain_id': domain_id})
        except Exception as e:
            _LOGGER.error(f'Failed to get managed policy. (policy_id = {policy_id})')
            return None

    def _get_managed_policy_from_local(self, policy_id, domain_id):
        managed_policy_vos = self.policy_model.filter(policy_id=policy_id, policy_type='MANAGED', domain_id=domain_id)
        if managed_policy_vos.count() > 0:
            return managed_policy_vos[0]
        else:
            return None

    def _delete_role_cache(self, policy_vo):
        role_vos = self._get_roles_using_policy(policy_vo)
        for role_vo in role_vos:
            cache.delete_pattern(f'role-permissions:*{role_vo.role_id}')
            cache.delete_pattern(f'user-permissions:*{role_vo.role_id}*')

    def _get_roles_using_policy(self, policy_vo):
        role_mgr: RoleManager = self.locator.get_manager('RoleManager')
        query = {
            'filter': [
                {
                    'k': 'policies.policy',
                    'v': policy_vo,
                    'o': 'eq'
                }
            ]
        }

        role_vos, total_count = role_mgr.list_roles(query)
        return role_vos



from spaceone.core.error import *


class ERROR_NOT_ALLOWED_ROLE_TYPE(ERROR_INVALID_ARGUMENT):
    _message = 'Duplicate assignment of system role and domain or project role is not allowed.'


class ERROR_REQUIRED_PROJECT_OR_PROJECT_GROUP(ERROR_INVALID_ARGUMENT):
    _message = 'Project role require project_id or project_group_id.'


class ERROR_NOT_ALLOWED_PROJECT_ID(ERROR_INVALID_ARGUMENT):
    _message = 'Domain or system role dose not need project_id.'


class ERROR_NOT_ALLOWED_PROJECT_GROUP_ID(ERROR_INVALID_ARGUMENT):
    _message = 'Domain or system role dose not need project_group_id.'


class ERROR_DUPLICATE_ROLE_BOUND(ERROR_INVALID_ARGUMENT):
    _message = 'Duplicate role bound. (role_id = {role_id}, resource_id = {resource_id})'


class ERROR_DUPLICATE_RESOURCE_IN_PROJECT(ERROR_INVALID_ARGUMENT):
    _message = 'There are duplicate resource in the project. (project_id = {project_id}, resource_id = {resource_id})'


class ERROR_DUPLICATE_RESOURCE_IN_PROJECT_GROUP(ERROR_INVALID_ARGUMENT):
    _message = 'There are duplicate resource in the project_group. (project_group_id = {project_group_id}, resource_id = {resource_id})'


class ERROR_POLICY_IS_IN_USE(ERROR_INVALID_ARGUMENT):
    _message = 'The policy is in use by role. (policy_id = {policy_id}, role_id = {role_id})'



import pytz
import random
import string
import re
from spaceone.core.service import *
from spaceone.core import config, utils
from spaceone.identity.error.error_user import *
from spaceone.identity.model import Domain, User
from spaceone.identity.manager import UserManager, DomainManager, DomainSecretManager, LocalTokenManager, EmailManager


@authentication_handler(exclude=['reset_password'])
@authorization_handler(exclude=['reset_password'])
@mutation_handler(exclude=['reset_password'])
@event_handler
class UserService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.user_mgr: UserManager = self.locator.get_manager('UserManager')

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['user_id', 'domain_id'])
    def create(self, params):
        """ Create user

        Args:
            params (dict): {
                'user_id': 'str',
                'password': 'str',
                'name': 'str',
                'email': 'str',
                'reset_password': 'bool',
                'user_type': 'str',
                'backend': 'str',
                'language': 'str',
                'timezone': 'str',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            user_vo (object)
        """

        params['user_type'] = params.get('user_type', 'USER')
        params['backend'] = params.get('backend', 'LOCAL')
        domain_id = params['domain_id']
        user_id = params['user_id']
        email = params.get('email')
        reset_password = params.get('reset_password', False)

        domain_mgr: DomainManager = self.locator.get_manager('DomainManager')
        domain_vo: Domain = domain_mgr.get_domain(domain_id)

        default_language = self._get_default_config(domain_vo, 'LANGUAGE')
        default_timezone = self._get_default_config(domain_vo, 'TIMEZONE')

        self._check_user_type_and_backend(params['user_type'], params['backend'], domain_vo)

        if 'language' not in params:
            params['language'] = default_language

        if 'timezone' not in params:
            params['timezone'] = default_timezone

        if 'timezone' in params:
            self._check_timezone(params['timezone'])

        if reset_password:
            self._check_reset_password_eligibility(user_id, params['backend'], email)

            email_manager: EmailManager = self.locator.get_manager('EmailManager')
            language = params['language']
            params['required_actions'] = ['UPDATE_PASSWORD']
            params['password'] = self._generate_temporary_password()

            reset_password_type = config.get_global('RESET_PASSWORD_TYPE')
            if reset_password_type == 'ACCESS_TOKEN':
                token = self._issue_temporary_token(user_id, domain_id)
                reset_password_link = self._get_console_sso_url(domain_id, token['access_token'])

                user_vo = self.user_mgr.create_user(params, domain_vo)
                email_manager.send_reset_password_email_when_user_added(user_id, email, reset_password_link, language)
            else:
                temp_password = params['password']
                console_link = self._get_console_url(domain_id)

                user_vo = self.user_mgr.create_user(params, domain_vo)
                email_manager.send_temporary_password_email_when_user_added(user_id, email, console_link, temp_password, language)
        else:
            user_vo = self.user_mgr.create_user(params, domain_vo)

        return user_vo

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['user_id', 'domain_id'])
    def update(self, params):
        """ Update user

        Args:
            params (dict): {
                'user_id': 'str',
                'password': 'str',
                'name': 'str',
                'email': 'str',
                'reset_password' : 'bool',
                'language': 'str',
                'timezone': 'str',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            user_vo (object)
        """

        if 'timezone' in params:
            self._check_timezone(params['timezone'])

        user_vo = self.user_mgr.get_user(params['user_id'], params['domain_id'])

        if params.get('reset_password'):
            domain_id = params['domain_id']
            user_id = user_vo.user_id
            backend = user_vo.backend
            email = params.get('email', user_vo.email)
            email_verified = user_vo.email_verified

            language = user_vo.language

            self._check_reset_password_eligibility(user_id, backend, email)

            if email_verified is False:
                raise ERROR_VERIFICATION_UNAVAILABLE(user_id=user_id)

            reset_password_type = config.get_global('RESET_PASSWORD_TYPE')
            email_manager: EmailManager = self.locator.get_manager('EmailManager')
            temp_password = self._generate_temporary_password()
            params['password'] = temp_password

            user_vo = self.user_mgr.update_user_by_vo(params, user_vo)
            user_vo = self.user_mgr.update_user_by_vo({'required_actions': ['UPDATE_PASSWORD']}, user_vo)

            if reset_password_type == 'ACCESS_TOKEN':
                token = self._issue_temporary_token(user_id, domain_id)
                reset_password_link = self._get_console_sso_url(domain_id, token['access_token'])

                email_manager.send_reset_password_email(user_id, email, reset_password_link, language)
            elif reset_password_type == 'PASSWORD':
                console_link = self._get_console_url(domain_id)

                email_manager.send_temporary_password_email(user_id, email, console_link, temp_password, language)
        else:
            user_vo = self.user_mgr.update_user_by_vo(params, user_vo)

        return user_vo

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['user_id', 'domain_id'])
    def verify_email(self, params):
        """ Verify email

        Args:
            params (dict): {
                'user_id': 'str',
                'email': 'str',
                'force': 'bool',
                'domain_id': 'str'
            }


        Returns:
            UserInfo
        """
        user_id = params['user_id']
        domain_id = params['domain_id']

        user_vo = self.user_mgr.get_user(user_id, domain_id)
        email = params.get('email', user_vo.email)
        force = params.get('force', False)

        if force:
            params['email_verified'] = True
            user_vo = self.user_mgr.update_user(params)
        else:
            params['email_verified'] = False
            user_vo = self.user_mgr.update_user(params)

            token_manager: LocalTokenManager = self.locator.get_manager('LocalTokenManager')
            verify_code = token_manager.create_verify_code(user_id, domain_id)

            email_manager: EmailManager = self.locator.get_manager('EmailManager')
            email_manager.send_verification_email(user_id, email, verify_code, user_vo.language)

        return user_vo

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['user_id', 'verify_code', 'domain_id'])
    def confirm_email(self, params):
        """ Confirm email

        Args:
            params (dict): {
                'user_id': 'str',
                'verify_code': 'str',
                'domain_id': 'str'
            }


        Returns:
            None
        """

        user_id = params['user_id']
        domain_id = params['domain_id']
        verify_code = params['verify_code']

        token_manager: LocalTokenManager = self.locator.get_manager('LocalTokenManager')

        if token_manager.check_verify_code(user_id, domain_id, verify_code):
            params['email_verified'] = True
            return self.user_mgr.update_user(params)
        else:
            raise ERROR_INVALID_VERIFY_CODE(verify_code=verify_code)

    @transaction
    @check_required(['user_id', 'domain_id'])
    def reset_password(self, params):
        """ Reset password

        Args:
            params (dict): {
                'user_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        user_vo: User = self.user_mgr.get_user(params['user_id'], params['domain_id'])
        user_id = params['user_id']
        domain_id = params['domain_id']
        backend = user_vo.backend
        email = user_vo.email
        language = user_vo.language

        self._check_reset_password_eligibility(user_id, backend, email)

        if user_vo.email_verified is False:
            raise ERROR_VERIFICATION_UNAVAILABLE(user_id=user_id)

        reset_password_type = config.get_global('RESET_PASSWORD_TYPE', 'ACCESS_TOKEN')
        email_manager: EmailManager = self.locator.get_manager('EmailManager')
        if reset_password_type == 'ACCESS_TOKEN':
            token = self._issue_temporary_token(user_id, domain_id)
            reset_password_link = self._get_console_sso_url(domain_id, token['access_token'])
            email_manager.send_reset_password_email(user_id, email, reset_password_link, language)

        elif reset_password_type == 'PASSWORD':
            temp_password = self._generate_temporary_password()
            self.user_mgr.update_user_by_vo({'password': temp_password}, user_vo)
            self.user_mgr.update_user_by_vo({'required_actions': ['UPDATE_PASSWORD']}, user_vo)
            console_link = self._get_console_url(domain_id)
            email_manager.send_temporary_password_email(user_id, email, console_link, temp_password, language)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['user_id', 'actions', 'domain_id'])
    def set_required_actions(self, params):
        """ Update user

        Args:
            params (dict): {
                'user_id': 'str',
                'actions': 'list',
                'domain_id': 'str'
            }

        Returns:
            user_vo (object)
        """

        new_actions = params['actions']

        user_vo: User = self.user_mgr.get_user(params['user_id'], params['domain_id'])

        if 'UPDATE_PASSWORD' in new_actions:
            if user_vo.backend == 'EXTERNAL' or user_vo.user_type == 'API_USER':
                raise ERROR_NOT_ALLOWED_ACTIONS(action='UPDATE_PASSWORD')

        required_actions = list(set(user_vo.required_actions + new_actions))

        return self.user_mgr.update_user_by_vo({'required_actions': required_actions}, user_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['user_id', 'domain_id'])
    def enable(self, params):
        """ Enable user

        Args:
            params (dict): {
                'user_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            user_vo (object)
        """

        return self.user_mgr.enable_user(params['user_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['user_id', 'domain_id'])
    def disable(self, params):
        """ Disable user

        Args:
            params (dict): {
                'user_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            user_vo (object)
        """

        return self.user_mgr.disable_user(params['user_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['user_id', 'domain_id'])
    def delete(self, params):
        """ Delete user

        Args:
            params (dict): {
                'user_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        return self.user_mgr.delete_user(params['user_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['search', 'domain_id'])
    def find(self, params):
        """ Disable user

        Args:
            params (dict): {
                'search' (one of): {
                    'user_id': 'str',
                    'keyword': 'str'
                },
                'domain_id': 'str'
            }

        Returns:
            results(list) : 'list of {
                                'user_id': 'str',
                                'name': 'str',
                                'email': 'str',
                                'tags': 'list'
                            }'
        """

        if not any(k in params['search'] for k in ['keyword', 'user_id']):
            raise ERROR_REQUIRED_PARAMETER(key='search.keyword | search.user_id')

        domain_mgr: DomainManager = self.locator.get_manager('DomainManager')
        domain_vo: Domain = domain_mgr.get_domain(params['domain_id'])

        # Check External Authentication from Domain
        if not domain_vo.plugin_info:
            raise ERROR_NOT_ALLOWED_EXTERNAL_AUTHENTICATION()

        return self.user_mgr.find_user(params['search'], domain_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['user_id', 'domain_id'])
    def get(self, params):
        """ Get user

        Args:
            params (dict): {
                'user_id': 'str',
                'domain_id': 'str',
                'only': 'list'
            }

        Returns:
            user_vo (object)
        """

        return self.user_mgr.get_user(params['user_id'], params['domain_id'], params.get('only'))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    @append_query_filter(['user_id', 'name', 'state', 'email', 'user_type', 'backend', 'role_id', 'domain_id'])
    @append_keyword_filter(['user_id', 'name', 'email'])
    def list(self, params):
        """ List users

        Args:
            params (dict): {
                'user_id': 'str',
                'name': 'str',
                'state': 'str',
                'email': 'str',
                'user_type': 'str',
                'backend': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            results (list): 'list of user_vo'
            total_count (int)
        """

        query: dict = params.get('query', {})
        return self.user_mgr.list_users(query)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id'])
    @append_keyword_filter(['user_id', 'name', 'email'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
            }

        Returns:
            values (list): 'list of statistics data'
            total_count (int)
        """

        query = params.get('query', {})
        return self.user_mgr.stat_users(query)

    def _get_domain_name(self, domain_id):
        domain_mgr: DomainManager = self.locator.get_manager('DomainManager')
        domain_vo: Domain = domain_mgr.get_domain(domain_id)
        return domain_vo.name

    def _issue_temporary_token(self, user_id, domain_id):
        identity_conf = config.get_global('identity') or {}
        timeout = identity_conf.get('temporary_token_timeout', 86400)

        domain_secret_mgr: DomainSecretManager = self.locator.get_manager('DomainSecretManager')
        private_jwk = domain_secret_mgr.get_domain_private_key(domain_id=domain_id)

        local_token_manager: LocalTokenManager = self.locator.get_manager('LocalTokenManager')
        return local_token_manager.issue_temporary_token(user_id, domain_id, private_jwk=private_jwk, timeout=timeout)

    def _get_console_sso_url(self, domain_id, token):
        domain_name = self._get_domain_name(domain_id)

        console_domain = config.get_global('EMAIL_CONSOLE_DOMAIN')
        console_domain = console_domain.format(domain_name=domain_name)

        return f'{console_domain}?sso_access_token={token}'

    def _get_console_url(self, domain_id):
        domain_name = self._get_domain_name(domain_id)

        console_domain = config.get_global('EMAIL_CONSOLE_DOMAIN')
        return console_domain.format(domain_name=domain_name)

    @staticmethod
    def _get_default_config(vo, item):
        DEFAULT = {
            'TIMEZONE': 'UTC',
            'LANGUAGE': 'en'
        }
        dict_domain = vo.to_dict()
        config = dict_domain.get('config', {})
        value = config.get(item, DEFAULT.get(item, None))
        return value

    @staticmethod
    def _check_timezone(timezone):
        if timezone not in pytz.all_timezones:
            raise ERROR_INVALID_PARAMETER(key='timezone', reason='Timezone is invalid.')

    @staticmethod
    def _check_user_type_and_backend(user_type, backend, domain_vo):
        # Check User Type and Backend
        if user_type == 'API_USER':
            if backend == 'EXTERNAL':
                raise ERROR_EXTERNAL_USER_NOT_ALLOWED_API_USER()

        # Check External Authentication from Domain
        if backend == 'EXTERNAL':
            if not domain_vo.plugin_info:
                raise ERROR_NOT_ALLOWED_EXTERNAL_AUTHENTICATION()

    @staticmethod
    def _check_reset_password_eligibility(user_id, backend, email):
        if backend == 'EXTERNAL':
            raise ERROR_UNABLE_TO_RESET_PASSWORD_IN_EXTERNAL_AUTH(user_id=user_id)
        elif email is None:
            raise ERROR_UNABLE_TO_RESET_PASSWORD_WITHOUT_EMAIL(user_id=user_id)

    @staticmethod
    def _generate_temporary_password():
        while True:
            random_password = ''.join(random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits) for _ in range(12))
            if re.search('[a-z]', random_password) and re.search('[A-Z]', random_password) and re.search('[0-9]', random_password):
                return random_password




import functools
from spaceone.api.identity.v1 import provider_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.identity.model.provider_model import Provider

__all__ = ['ProviderInfo', 'ProvidersInfo']


def ProviderInfo(provider_vo: Provider, minimal=False):
    info = {
        'provider': provider_vo.provider,
        'name': provider_vo.name
    }

    if not minimal:
        info.update({
            'template': change_struct_type(provider_vo.template),
            'metadata': change_struct_type(provider_vo.metadata),
            'capability': change_struct_type(provider_vo.capability),
            'tags': change_struct_type(provider_vo.tags),
            'created_at': utils.datetime_to_iso8601(provider_vo.created_at)
        })

    return provider_pb2.ProviderInfo(**info)


def ProvidersInfo(provider_vos, total_count, **kwargs):
    results = list(map(functools.partial(ProviderInfo, **kwargs), provider_vos))

    return provider_pb2.ProvidersInfo(results=results, total_count=total_count)



from spaceone.api.identity.v1 import domain_owner_pb2, domain_owner_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class DomainOwner(BaseAPI, domain_owner_pb2_grpc.DomainOwnerServicer):

    pb2 = domain_owner_pb2
    pb2_grpc = domain_owner_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DomainOwnerService', metadata) as domain_owner_svc:
            data = domain_owner_svc.create(params)
            return self.locator.get_info('DomainOwnerInfo', data)

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DomainOwnerService', metadata) as domain_owner_svc:
            data = domain_owner_svc.update(params)
            return self.locator.get_info('DomainOwnerInfo', data)

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DomainOwnerService', metadata) as domain_owner_svc:
            # TODO: block delete when users, projects, inventories are existing.
            data = domain_owner_svc.delete(params)
            return self.locator.get_info('EmptyInfo')

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('DomainOwnerService', metadata) as domain_owner_svc:
            data = domain_owner_svc.get(params)
            return self.locator.get_info('DomainOwnerInfo', data)



from spaceone.api.identity.v1 import provider_pb2, provider_pb2_grpc
from spaceone.core.pygrpc import BaseAPI


class Provider(BaseAPI, provider_pb2_grpc.ProviderServicer):

    pb2 = provider_pb2
    pb2_grpc = provider_pb2_grpc

    def create(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProviderService', metadata) as provider_svc:
            data = provider_svc.create(params)
            return self.locator.get_info('ProviderInfo', data)

    def update(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProviderService', metadata) as provider_svc:
            data = provider_svc.update(params)
            return self.locator.get_info('ProviderInfo', data)

    def delete(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProviderService', metadata) as provider_svc:
            provider_svc.delete(params)
            return self.locator.get_info('EmptyInfo')

    def get(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProviderService', metadata) as provider_svc:
            data = provider_svc.get(params)
            return self.locator.get_info('ProviderInfo', data)

    def list(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProviderService', metadata) as provider_svc:
            provider_vos, total_count = provider_svc.list(params)
            return self.locator.get_info('ProvidersInfo', provider_vos,
                                         total_count, minimal=self.get_minimal(params))

    def stat(self, request, context):
        params, metadata = self.parse_request(request, context)

        with self.locator.get_service('ProviderService', metadata) as provider_svc:
            return self.locator.get_info('StatisticsInfo', provider_svc.stat(params))



import functools
from spaceone.api.identity.v1 import project_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.identity.model.project_model import Project
from spaceone.identity.info.project_group_info import ProjectGroupInfo
from spaceone.identity.info.role_info import RoleInfo

__all__ = ['ProjectInfo', 'ProjectsInfo', 'ProjectRoleBindingInfo', 'ProjectRoleBindingsInfo']


def ProjectInfo(project_vo: Project, minimal=False, project_groups_info=None):
    info = {
        'project_id': project_vo.project_id,
        'name': project_vo.name

    }

    if not minimal:
        if project_groups_info:
            if project_vo.project_group_id and project_vo.project_group_id in project_groups_info:
                project_group_info = ProjectGroupInfo(project_groups_info[project_vo.project_group_id], minimal=True)
                info.update({
                    'project_group_info': project_group_info
                })
        else:
            if project_vo.project_group:
                info.update({
                    'project_group_info': ProjectGroupInfo(project_vo.project_group, minimal=True)
                })

        info.update({
            'tags': change_struct_type(project_vo.tags),
            'domain_id': project_vo.domain_id,
            'created_by': project_vo.created_by,
            'created_at': utils.datetime_to_iso8601(project_vo.created_at)
        })

    return project_pb2.ProjectInfo(**info)


def ProjectsInfo(project_vos, total_count, **kwargs):
    results = list(map(functools.partial(ProjectInfo, **kwargs), project_vos))
    return project_pb2.ProjectsInfo(results=results, total_count=total_count)


def ProjectRoleBindingInfo(role_binding_vo, minimal=False):
    info = {
        'role_binding_id': role_binding_vo.role_binding_id,
        'resource_type': role_binding_vo.resource_type,
        'resource_id': role_binding_vo.resource_id,
        'role_info': RoleInfo(role_binding_vo.role, minimal=True) if role_binding_vo.role else None
    }

    if not minimal:
        info.update({
            'project_info': ProjectInfo(role_binding_vo.project, minimal=True) if role_binding_vo.project else None,
            'project_group_info': ProjectGroupInfo(role_binding_vo.project_group, minimal=True) if role_binding_vo.project_group else None,
            'labels': change_list_value_type(role_binding_vo.labels),
            'tags': change_struct_type(role_binding_vo.tags),
            'domain_id': role_binding_vo.domain_id,
            'created_at': utils.datetime_to_iso8601(role_binding_vo.created_at)
        })

    return project_pb2.ProjectRoleBindingInfo(**info)


def ProjectRoleBindingsInfo(role_binding_vos, total_count, **kwargs):
    results = list(map(functools.partial(ProjectRoleBindingInfo, **kwargs), role_binding_vos))

    return project_pb2.ProjectRoleBindingsInfo(results=results, total_count=total_count)



DEFAULT_PROVIDERS = [{
    "provider": "aws",
    "name": "AWS",
    "template": {
        "service_account": {
            "schema": {
                "type": "object",
                "properties": {
                    "account_id": {
                        "title": "Account ID",
                        "type": "string",
                        "minLength": 4
                    }
                },
                "required": ["account_id"]
            }
        }
    },
    "metadata": {
        "view": {
            "layouts": {
                "help:service_account:create": {
                    "name": "Creation Help",
                    "type": "markdown",
                    "options": {
                        "markdown": {
                            "en": (
                                "# Help for AWS Users\n"
                                "## Find Your AWS Account ID\n"
                                "Get your AWS Account ID.\n"
                                "[AWS Account ID](https://docs.aws.amazon.com/IAM/latest/UserGuide/console_account-alias.html)\n"
                                "## Get Your Assume role\n"
                                "Granting permissions to create temporary security credentials.\n"
                                "[AWS Assume Role](https://docs.aws.amazon.com/IAM/latest/UserGuide/id_roles_use_permissions-to-switch.html)\n"
                                "## Issue AWS Access Key \n"
                                "Get your AWS Access Key & AWS Secret Key\n"
                                "[AWS Access Key & AWS Secret Key](https://docs.aws.amazon.com/IAM/latest/UserGuide/id_credentials_access-keys.html#Using_CreateAccessKey)\n"
                            ),
                            "ko": (
                                "# AWS 이용자 가이드\n"
                                "## AWS 어카운트 아이디(Account ID) 찾기\n"
                                "사용자의 AWS 어카운트 아이디 AWS 콘솔(Console)에서 확인하기\n"
                                "[AWS Account ID](https://docs.aws.amazon.com/ko_kr/IAM/latest/UserGuide/console_account-alias.html)\n"
                                "## Assume role 획득하기\n"
                                "임시 보안 자격증명을 만들 수있는 권한을 부여하기.\n"
                                "[AWS Assume Role](https://docs.aws.amazon.com/ko_kr/IAM/latest/UserGuide/id_roles_use_permissions-to-switch.html)\n"
                                "## AWS Access Key 발급하기\n"
                                "AWS Access Key & AWS Secret Key 발급하기\n"
                                "[AWS Access Key & AWS Secret Key](https://docs.aws.amazon.com/ko_kr/IAM/latest/UserGuide/id_credentials_access-keys.html#Using_CreateAccessKey)\n"
                            ),

                        }
                    }
                }
            }
        }
    },
    "capability": {
        "general_service_account_schema": [
            "aws_assume_role"
        ],
        "trusted_service_account_schema": [
            "aws_access_key"
        ],
        "support_trusted_service_account": True,
        "supported_schema": [
            "aws_access_key"
        ]
    },
    "tags": {
        'color': '#FF9900',
        'icon': 'https://spaceone-custom-assets.s3.ap-northeast-2.amazonaws.com/console-assets/icons/aws.svg',
        'external_link_template': 'https://<%- data.account_id %>.signin.aws.amazon.com/console'
    }
}, {
    "provider": "google_cloud",
    "version": "v1",
    "name": "Google Cloud",
    "template": {
        "service_account": {
            "schema": {
                "type": "object",
                "properties": {
                    "project_id": {
                        "title": "Project ID",
                        "type": "string",
                        "minLength": 4
                    }
                },
                "required": ["project_id"]
            }
        }
    },
    "metadata": {
        "view": {
            "layouts": {
                "help:service_account:create": {
                    "name": "Creation Help",
                    "type": "markdown",
                    "options": {
                        "markdown": {
                            "en": (
                                "# Getting started with Google Cloud\n"
                                "## Identifying Your Project\n"
                                "Get your Project infos (Project Name, Project ID and Project number)\n"
                                "[Project Info](https://cloud.google.com/resource-manager/docs/creating-managing-projects#identifying_projects)\n"
                                "## Get Your Service Account Key(JSON)\n"
                                "Generate Your JSON Service Account Key.\n"
                                "[Service Account Key](https://cloud.google.com/docs/authentication/getting-started)\n"
                            ),
                            "ko": (
                                "# Google Cloud 시작 가이드\n"
                                "## Project 정보 확인하기\n"
                                "프로젝트 명, 프로젝트 아이디, 프로젝트 번호 등의 프로젝트 정보 확인하기\n"
                                "[Project Info](https://cloud.google.com/resource-manager/docs/creating-managing-projects?hl=ko#identifying_projects)\n"
                                "## 서비스 어카운트 키(JSON) 받기\n"
                                "JSON 포멧의 서비스 어카운트 키를 생성하기.\n"
                                "[Service Account Key](https://cloud.google.com/docs/authentication/getting-started?hl=ko)\n"
                            ),

                        }
                    }
                }
            }
        }
    },
    "capability": {
        "general_service_account_schema": [
            "google_cloud_project_id"
        ],
        "trusted_service_account_schema": [
            "google_oauth2_credentials"
        ],
        "support_trusted_service_account": True,
        "supported_schema": ["google_oauth2_credentials"]
    },
    "tags": {
        'color': '#4285F4',
        'icon': 'https://spaceone-custom-assets.s3.ap-northeast-2.amazonaws.com/console-assets/icons/google_cloud.svg',
        'external_link_template': 'https://console.cloud.google.com/home/dashboard?project=<%- data.project_id %>',
        'label': 'Google'
    }
}, {
    "provider": "azure",
    "name": "Azure",
    "template": {
        "service_account": {
            "schema": {
                "type": "object",
                "properties": {
                    "tenant_id": {
                        "title": "Tenant ID",
                        "type": "string",
                        "minLength": 4
                    },
                    "subscription_id": {
                        "title": "Subscription ID",
                        "type": "string",
                        "minLength": 4
                    }
                },
                "required": ["tenant_id", "subscription_id"]
            }
        }
    },
    "metadata": {
        "view": {
            "layouts": {
                "help:service_account:create": {
                    "name": "Creation Help",
                    "type": "markdown",
                    "options": {
                        "markdown": {
                            "en": (
                                "# Help for Azure Users\n"
                                "## Find Your Azure Subscription ID\n"
                                "Azure Subscription ID via CLI.\n"
                                "[Azure Subscription CLI](https://docs.microsoft.com/en-us/cli/azure/ext/account/account/subscription?view=azure-cli-latest)\n"
                                "Azure Subscription ID via PowerShell.\n"
                                "[Azure Subscription PowerShell](https://docs.microsoft.com/en-us/azure/active-directory/develop/howto-authenticate-service-principal-powershell)\n"
                                "Create Azure Subscription via Portal.\n"
                                "[Azure Subscription Portal](https://docs.microsoft.com/en-us/azure/active-directory/develop/howto-create-service-principal-portal)\n"
                                "## Find Your Azure Tenant ID\n"
                                "Azure Tenant ID via CLI.\n"
                                "[Azure Tenant CLI](https://docs.microsoft.com/en-us/cli/azure/ext/account/account/tenant?view=azure-cli-latest)\n"
                                "Azure Tenant ID via PowerShell.\n"
                                "[Azure Tenant PowerShell](https://docs.microsoft.com/en-us/powershell/module/az.accounts/get-aztenant?view=azps-5.0.0)\n"
                                "## Get Your Client Secret and ID\n"
                                "Check Client Secret via Portal.\n"
                                "[Azure Client Secret Portal](https://docs.microsoft.com/en-us/azure/active-directory/develop/howto-create-service-principal-portal)\n"
                            ),
                            "ko": (
                                "# Azure 이용자 가이드\n"
                                "## Azure 구독 아이디(Subscription ID) 찾기\n"
                                "CLI에서 사용자의 구독 아이디 확인하기.\n"
                                "[Azure Subscription CLI](https://docs.microsoft.com/en-us/cli/azure/ext/account/account/subscription?view=azure-cli-latest)\n"
                                "PowerShell에서 사용자의 구독 아이디 확인하기.\n"
                                "[Azure Subscription PowerShell](https://docs.microsoft.com/en-us/azure/active-directory/develop/howto-authenticate-service-principal-powershell)\n"
                                "포털에서 사용자의 구독 아이디 확인하기.\n"
                                "[Azure Subscription Portal](https://docs.microsoft.com/en-us/azure/active-directory/develop/howto-create-service-principal-portal)\n"
                                "## Azure 테넌트 아이디(Tenant ID) 찾기\n"
                                "CLI에서 사용자의 테넌트 아이디 확인하기.\n"
                                "[Azure Tenant CLI](https://docs.microsoft.com/en-us/cli/azure/ext/account/account/tenant?view=azure-cli-latest)\n"
                                "PowerShell에서 사용자의 테넌트 아이디 확인하기.\n"
                                "[Azure Tenant PowerShell](https://docs.microsoft.com/en-us/powershell/module/az.accounts/get-aztenant?view=azps-5.0.0)\n"
                                "## 사용자의 클라이언트 시크릿 정보(Client Secret&ID) 가져오기\n"
                                "포털에서 사용자의 클라이언트 시크릿 정보 확인하기.\n"
                                "[Azure Client Secret Portal](https://docs.microsoft.com/en-us/azure/active-directory/develop/howto-create-service-principal-portal)\n"
                            ),

                        }
                    }
                }
            }
        }
    },
    "capability": {
        "general_service_account_schema": [
            "azure_subscription_id"
        ],
        "trusted_service_account_schema": [
            "azure_credentials",
        ],
        "support_trusted_service_account": True,
        "supported_schema": ["azure_client_secret"]
    },
    "tags": {
        'color': '#00BCF2',
        'icon': 'https://spaceone-custom-assets.s3.ap-northeast-2.amazonaws.com/console-assets/icons/azure.svg'
    }
}]

# }, {
#     "provider": "spaceone",
#     "name": "SpaceONE",
#     "template": {
#         "service_account": {
#             "schema": {
#                 "type": "object",
#                 "properties": {
#                     "user_id": {
#                         "title": "User ID",
#                         "type": "string",
#                         "minLength": 4
#                     }
#                 },
#                 "required": ["user_id"]
#             }
#         }
#     },
#     "capability": {
#         "supported_schema": ["spaceone_api_key"]
#     },
#     "tags": [
#         {
#             'key': 'color',
#             'value': '#6638B6'
#         }, {
#             'key': 'icon',
#             'value': 'https://spaceone-custom-assets.s3.ap-northeast-2.amazonaws.com/console-assets/icons/spaceone.svg'
#         }
#     ]
# }, {
#     "provider": "openstack",
#     "name": "OpenStack",
#     "template": {
#         "service_account": {
#             "schema": {
#                 "type": "object",
#                 "properties": {
#                     "username": {
#                         "title": "Description",
#                         "type": "string",
#                         "minLength": 4
#                     }
#                 }
#             }
#         }
#     },
#     "metadata": {
#         "view": {
#             "layouts": {
#             }
#         }
#     },
#     "capability": {
#         "supported_schema": ["openstack_credentials"]
#     },
#     "tags": [
#         {
#             'key': 'color',
#             'value': '#EE1142'
#         }, {
#             'key': 'icon',
#             'value': 'https://spaceone-custom-assets.s3.ap-northeast-2.amazonaws.com/console-assets/icons/openstack.svg'
#         }
#     ]
# }, {
#     "provider": "alibaba_cloud",
#     "name": "Alibaba Cloud",
#     "template": {
#         "service_account": {
#             "schema": {
#                 "type": "object",
#                 "properties": {
#                     "account_id": {
#                         "title": "Account ID",
#                         "type": "string",
#                         "minLength": 4
#                     }
#                 },
#                 "required": ["account_id"]
#             }
#         }
#     },
#     "metadata": {
#         "view": {
#             "layouts": {
#                 "help:service_account:create": {
#                     "name": "Creation Help",
#                     "type": "markdown",
#                     "options": {
#                         "markdown": {
#                             "en": (
#                                 "# Help for Alibaba Cloud Users\n"
#                                 "## Find Your Alibaba Cloud Account ID\n"
#                                 "Get your Alibaba Cloud Account ID.\n"
#                                 "Individual Account: Alibaba Cloud console > Click your avatar > Account Management Page > Security Settings > **Account ID**\nEnterprise Account: Alibaba Cloud console > Click your avatar > **Enterprise Alias**\n"
#                                 "## Generate Your AccessKey Pair\n"
#                                 "Get your Alibaba Cloud AccessKey ID & AccessKey Secret\n"
#                                 "[Alibaba Cloud AccessKey & AccessSecret](https://www.alibabacloud.com/help/doc-detail/53045.htm?spm=a2c63.p38356.879954.17.6be510df6F9vxS#concept-53045-zh)\n"
#                             ),
#                             "ko": (
#                                 "# 알리바바 클라우드 이용자 가이드\n"
#                                 "## 알리바바 클라우드 계정 아이디(Account ID) 찾기\n"
#                                 "콘솔에서 사용자의 알리바바 클라우드 계정 아이디 확인하기\n"
#                                 "개인 계정: 알리바바 클라우드 콘솔 > 사용자 아바타 클릭 > Account Management 페이지 > Security Settings > **Account ID**\n엔터프라이즈 계정: 알리바바 클라우드 콘솔 > 사용자 아바타 클 > **Enterprise Alias**\n"
#                                 "## 알리바바 클라우드 AccessKey Pair 발급하기\n"
#                                 "콘솔에서 알리바바 클라우드 AccessKey ID & AccessKey Secret 생성하기\n"
#                                 "[Alibaba Cloud AccessKey & AccessKey Secret](https://www.alibabacloud.com/help/doc-detail/53045.htm?spm=a2c63.p38356.879954.17.6be510df6F9vxS#concept-53045-zh)\n"
#                             ),
#                         }
#                     }
#                 }
#             }
#         }
#     },
#     "capability": {
#         "supported_schema": ["alibaba_cloud_access_key_pair"]
#     },
#     "tags": [
#         {
#             'key': 'color',
#             'value': '#FF6A00'
#         }, {
#             'key': 'icon',
#             'value': 'https://spaceone-custom-assets.s3.ap-northeast-2.amazonaws.com/console-assets/icons/cloud-services/alibaba_cloud/logo.svg'
#         }, {
#             'key': 'external_link_template',
#             'value': 'https://homenew-intl.console.aliyun.com/'
#         }
#     ]
# }, {
#     "provider": "oracle_cloud",
#     "name": "Oracle Cloud Infrastructure",
#     "template": {
#         "service_account": {
#             "schema": {
#                 "type": "object",
#                 "properties": {
#                     "tenancy_name": {
#                         "title": "Tenancy Name",
#                         "type": "string",
#                         "minLength": 4
#                     },
#                     "user_name": {
#                         "title": "User Name",
#                         "type": "string",
#                         "minLength": 4
#                     }
#                 },
#                 "required": ["tenancy_name", "user_name"]
#             }
#         }
#     },
#     "metadata": {
#         "view": {
#             "layouts": {
#                 "help:service_account:create": {
#                     "name": "Creation Help",
#                     "type": "markdown",
#                     "options": {
#                         "markdown": {
#                             "en": (
#                                 "# Getting started with Oracle Cloud Intrastructure\n"
#                                 "## Find Your Tenancy & User OCID\n"
#                                 "[Get your User's OCID  and Tenancy's OCID](https://docs.oracle.com/en-us/iaas/Content/API/Concepts/apisigningkey.htm#five)\n"
#                                 "## Generate RSA key pair and Key's Fingerprint\n"
#                                 "[Generate RSA key pair]([https://docs.oracle.com/en-us/iaas/Content/API/Concepts/apisigningkey.htm#two](https://docs.oracle.com/en-us/iaas/Content/API/Concepts/apisigningkey.htm#two))\n"
#                                 "[Get Key's Fingerprint](https://docs.oracle.com/en-us/iaas/Content/API/Concepts/apisigningkey.htm#four)\n"
#                                 "## Upload the Public Key from the key pair in the Console\n"
#                                 "[Upload Your Public Key](https://docs.oracle.com/en-us/iaas/Content/API/Concepts/apisigningkey.htm#three)\n"
#                                 #R"## Note: It requires to add \n at end of every each line of your private_key string prior to put private key in credentials. ex) -----BEGIN PRIVATE KEY-----\nMIIEvQIBADANBgkqhkiG9w0\n...\n..\n"
#                                 "## Note: You must copy and paste the private key, except for the header and footer of the private_key string, before putting it into Credential. The same is true for typing in JSON format; each string is separated by a space.\n"
#                             ),
#                             "ko": (
#                                 "# 오라클 클라우드 이용자 가이드\n"
#                                 "## 오라클 클라우드 테넌시와 유저 OCID 찾기\n"
#                                 "[유저와 테넌시 OCID 정보 확인](https://docs.oracle.com/en-us/iaas/Content/API/Concepts/apisigningkey.htm#five)\n"
#                                 "## RSA 키 페어와 핑거프린트 생성\n"
#                                 "[RSA 키 페어 생성]([https://docs.oracle.com/en-us/iaas/Content/API/Concepts/apisigningkey.htm#two](https://docs.oracle.com/en-us/iaas/Content/API/Concepts/apisigningkey.htm#two))\n"
#                                 "[핑거프린트 생성](https://docs.oracle.com/en-us/iaas/Content/API/Concepts/apisigningkey.htm#four)\n"
#                                 "## 오라클 클라우드 콘솔에 공개 키 등록\n"
#                                 "[퍼블릭 키 업로드](https://docs.oracle.com/en-us/iaas/Content/API/Concepts/apisigningkey.htm#three)\n"
#                                 "## Note: 개인 키를 Credential에 넣기 전에 private_key 문자열의 헤더와 푸터를 제외하고 복사해서 붙여넣어야 합니다. JSON 형식으로 입력 시에도 동일하며, 각 문자열은 공백으로 구분합니다.\n  "
#                             ),
#                         }
#                     }
#                 }
#             }
#         }
#     },
#     "capability": {
#         "supported_schema": ["oci_api_key"]
#     },
#     "tags": [
#         {
#             'key': 'color',
#             'value': '#D73A27'
#         }, {
#             'key': 'icon',
#             'value': 'https://spaceone-custom-assets.s3.ap-northeast-2.amazonaws.com/console-assets/icons/cloud-services/oci/ic_provider_oracle.svg'
#         }, {
#             'key': 'external_link_template',
#             'value': 'https://www.oracle.com/kr/cloud/sign-in.html'
#         }
#     ]

aws_access_key = {
    "name": "aws_access_key",
    "service_type": "secret.credentials",
    "schema": {
        "required": [
            "aws_access_key_id",
            "aws_secret_access_key"
        ],
        "order": [
            "aws_access_key_id",
            "aws_secret_access_key"
        ],
        "properties": {
            "aws_access_key_id": {
                "title": "AWS Access Key",
                "type": "string",
                "format": "password",
                "minLength": 4
            },
            "aws_secret_access_key": {
                "title": "AWS Secret Key",
                "type": "string",
                "format": "password",
                "minLength": 4
            }
        },
        "type": "object"
    },
    "labels": ["AWS"],
    "tags": {"description": "AWS Access Key"}
}

aws_assume_role = {
    "name": "aws_assume_role",
    "service_type": "secret.credentials",
    "schema": {
        "required": [
            "external_id",
            "role_arn"
        ],
        "order": [
            "external_id",
            "role_arn"
        ],
        "properties": {
            "external_id": {
                "minLength": 4.0,
                "title": "External ID",
                "markdown": "[How to use an external ID?](https://docs.aws.amazon.com/IAM/latest/UserGuide/id_roles_create_for-user_externalid.html)",
                "type": "string",
                "format": "generate_id"
            },
            "role_arn": {
                "type": "string",
                "minLength": 4.0,
                "title": "Role ARN"
            }
        },
        "type": "object"
    },
    "labels": ["AWS", "Assume Role"],
    "tags": {"description": "AWS Assume Role"}
}

google_oauth2_credentials = {
    "name": "google_oauth2_credentials",
    "service_type": "secret.credentials",
    "schema": {
        "required": [
            "type",
            "project_id",
            "private_key_id",
            "private_key",
            "client_email",
            "client_id",
            "auth_uri",
            "token_uri",
            "auth_provider_x509_cert_url",
            "client_x509_cert_url"
        ],
        "properties": {
            "client_x509_cert_url": {
                "examples": [
                    "https://www.googleapis.com/..."
                ],
                "minLength": 4.0,
                "type": "string",
                "title": "Client X509 Cert URL"
            },
            "auth_uri": {
                "title": "Auth URI",
                "minLength": 4.0,
                "default": "https://acounts.google.com/o/oauth2/auth",
                "type": "string"
            },
            "private_key_id": {
                "examples": [
                    "771823abcd..."
                ],
                "minLength": 4.0,
                "type": "string",
                "title": "Private Key ID"
            },
            "zone": {
                "examples": [
                    "asia-northeast3"
                ],
                "type": "string",
                "title": "Region",
                "minLength": 4.0
            },
            "type": {
                "minLength": 4.0,
                "type": "string",
                "default": "service_account",
                "title": "Type"
            },
            "client_email": {
                "title": "Client Email",
                "examples": [
                    "<api-name>api@project-id.iam.gserviceaccount.com(opens in new tab)"
                ],
                "type": "string",
                "minLength": 4.0
            },
            "auth_provider_x509_cert_url": {
                "title": "Auth Provider Cert URL",
                "minLength": 4.0,
                "default": "https://www.googleapis.com/oauth2/v1/certs",
                "type": "string"
            },
            "private_key": {
                "minLength": 4.0,
                "examples": [
                    "-----BEGIN"
                ],
                "type": "string",
                "title": "Private Key"
            },
            "token_uri": {
                "title": "Token URI",
                "minLength": 4.0,
                "default": "https://oauth2.googleapis.com/token",
                "type": "string"
            },
            "project_id": {
                "minLength": 4.0,
                "title": "Project ID",
                "examples": [
                    "project-id"
                ],
                "type": "string"
            },
            "client_id": {
                "examples": [
                    "10118252....."
                ],
                "minLength": 4.0,
                "type": "string",
                "title": "Client ID"
            }
        },
        "type": "object"
    },
    "labels": ["Google Cloud", "GCP", "OAuth2.0"],
    "tags": {"description": "Google OAuth2 Credentials"}
}

azure_client_secret = {
    "name": "azure_client_secret",
    "service_type": "secret.credentials",
    "schema": {
        "required": [
            "subscription_id",
            "tenant_id",
            "client_id",
            "client_secret"
        ],
        "order": [
            "subscription_id",
            "tenant_id",
            "client_id",
            "client_secret"
        ],
        "properties": {
            "client_id": {
                "title": "Client ID",
                "type": "string",
                "minLength": 4
            },
            "client_secret": {
                "title": "Client Secret",
                "type": "string",
                "format": "password",
                "minLength": 4
            },
            "tenant_id": {
                "title": "Tenant ID",
                "type": "string",
                "minLength": 4
            },
            "subscription_id": {
                "title": "Subscription ID",
                "type": "string",
                "minLength": 4
            }
        },
        "type": "object"
    },
    "labels": ["Azure"],
    "tags": {"description": "Azure Client Secret"}
}

spaceone_api_key = {
    "name": "spaceone_api_key",
    "service_type": "secret.credentials",
    "schema": {
        "required": [
            "api_key",
            "endpoint"
        ],
        "order": [
            "api_key",
            "endpoint"
        ],
        "properties": {
            "api_key": {
                "title": "API Key",
                "type": "string",
                "format": "password",
                "minLength": 4
            },
            "endpoint": {
                "title": "Identity Service Endpoint",
                "type": "string",
                "minLength": 4
            }
        },
        "type": "object"
    },
    "labels": ["SpaceONE"],
    "tags": {"description": "SpaceONE API Key"}
}

alibaba_cloud_access_key_pair = {
    "name": "alibaba_cloud_access_key_pair",
    "service_type": "secret.credentials",
    "schema": {
        "required": [
            "ali_access_key_id",
            "ali_access_key_secret"
        ],
        "properties": {
            "ali_access_key_id": {
                "title": "Alibaba Cloud AccessKey ID",
                "type": "string",
                "format": "password",
                "minLength": 4
            },
            "ali_access_key_secret": {
                "title": "Alibaba Cloud AccessKey Secret",
                "type": "string",
                "format": "password",
                "minLength": 4
            }
        },
        "type": "object"
    },
    "labels": ["Alibaba"],
    "tags": {"description": "Alibaba Cloud AccessKey Pair"}
}

oci_api_key = {
    "name": "oci_api_key",
    "service_type": "secret.credentials",
    "schema": {
        "required": [
            "tenancy",
            "user",
            "key_content",
            "fingerprint"
        ],
        "properties": {
            "tenancy": {
                "title": "Tenancy's OCID",
                "type": "string",
                "minLength": 4
            },
            "user": {
                "title": "User's OCID",
                "type": "string",
                "minLength": 4
            },
            "key_content": {
                "title": "Private Key Content",
                "type": "string",
                "minLength": 4
            },
            "fingerprint": {
                "title": "Fingerprint",
                "type": "string",
                "minLength": 4
            }
        },
        "type": "object"
    },
    "labels": ["Oracle"],
    "tags": {"description": "Oracle Cloud Infrastructure Access Configuration"}
}

openstack_credentials = {
    "name": "openstack_credentials",
    "service_type": "secret.credentials",
    "schema": {
        "required": [
            "username",
            "password",
            "region_name",
            "auth_url",
            "project_id"
        ],
        "properties": {
            "username": {
                "title": "Username",
                "type": "string",
                "minLength": 1
            },
            "password": {
                "title": "Password",
                "type": "string",
                "minLength": 1
            },
            "region_name": {
                "title": "Region",
                "type": "string",
                "minLength": 1
            },
            "interface": {
                "title": "API Interface",
                "type": "string",
                "minLength": 1,
                "examples": ["public"]
            },
            "identity_api_version": {
                "title": "Identity API Version",
                "type": "string",
                "minLength": 1,
                "examples": ["3"]
            },
            "auth_url": {
                "title": "Auth URL",
                "type": "string",
                "minLength": 1
            },
            "user_domain_name": {
                "title": "Domain",
                "type": "string",
                "minLength": 1,
                "examples": ["Default"]
            },
            "project_id": {
                "title": "Project ID",
                "type": "string",
                "minLength": 1
            },
            "dashboard_url": {
                "title": "Dashboard URL",
                "type": "string",
                "minLength": 1
            },
            "all_projects": {
                "title": "All Projects",
                "type": "string",
                "enum": ["true", "false"],
                "minLength": 4,
                "examples": ["true"]
            }
        },
        "type": "object"
    },
    "labels": ["OpenStack"],
    "tags": {"description": "OpenStack Credentials"}
}



import functools
import json

from spaceone.api.core.v1 import handler_pb2
from spaceone.api.identity.v1 import domain_pb2

from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.identity.model.domain_model import Domain

__all__ = ['DomainInfo', 'DomainsInfo', 'DomainPublicKeyInfo']


def DomainInfo(domain_vo: Domain, minimal=False):
    info = {
        'domain_id': domain_vo.domain_id,
        'name': domain_vo.name,
        'state': domain_vo.state
    }

    if not minimal:
        info.update({
            'plugin_info': PluginInfo(domain_vo.plugin_info),
            'config': change_struct_type(domain_vo.config),
            'created_at': utils.datetime_to_iso8601(domain_vo.created_at),
            'deleted_at': utils.datetime_to_iso8601(domain_vo.deleted_at),
            'tags': change_struct_type(domain_vo.tags)
        })

    return domain_pb2.DomainInfo(**info)


def DomainsInfo(domain_vos, total_count, **kwargs):
    results = list(map(functools.partial(DomainInfo, **kwargs), domain_vos))

    return domain_pb2.DomainsInfo(results=results, total_count=total_count)


def PluginInfo(plugin_info):
    if plugin_info:
        info = {
            'plugin_id': plugin_info.plugin_id,
            'version': plugin_info.version,
            'options': change_struct_type(plugin_info.options),
            'metadata': change_struct_type(plugin_info.metadata),
            'secret_id': plugin_info.secret_id,
            'schema': plugin_info.schema,
            'upgrade_mode': plugin_info.upgrade_mode
        }
        return domain_pb2.PluginInfo(**info)

    return None


def DomainPublicKeyInfo(public_key, domain_id):
    info = {
        'public_key': json.dumps(public_key).__str__(),
        'domain_id': domain_id
    }
    return handler_pb2.AuthenticationResponse(**info)



import logging
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from spaceone.core.connector import BaseConnector

__all__ = ['SMTPConnector']

_LOGGER = logging.getLogger(__name__)


class SMTPConnector(BaseConnector):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.smtp = None
        host = self.config.get('host')
        port = self.config.get('port')
        user = self.config.get('user')
        password = self.config.get('password')
        self.from_email = self.config.get('from_email')
        self.set_smtp(host, port, user, password)

    def set_smtp(self, host, port, user, password):
        self.smtp = smtplib.SMTP(host, port)
        self.smtp.connect(host, port)
        self.smtp.ehlo()
        self.smtp.starttls()
        self.smtp.login(user, password)

    def send_email(self, to_emails, subject, contents):
        multipart_msg = MIMEMultipart("alternative")

        multipart_msg["Subject"] = subject
        multipart_msg["From"] = self.from_email
        multipart_msg["To"] = to_emails

        multipart_msg.attach(MIMEText(contents, 'html'))

        self.smtp.sendmail(self.from_email, to_emails.split(','), multipart_msg.as_string())

    def quit_smtp(self):
        self.smtp.quit()



from spaceone.api.identity.v1 import domain_owner_pb2

from spaceone.core import utils
from spaceone.identity.model import DomainOwner

__all__ = ['DomainOwnerInfo']


def DomainOwnerInfo(owner_vo: DomainOwner, minimal=False):
    info = {
        'owner_id': owner_vo.owner_id,
        'name': owner_vo.name
    }

    if not minimal:
        info.update({
            'email': owner_vo.email,
            'language': owner_vo.language,
            'timezone': owner_vo.timezone,
            'domain_id': owner_vo.domain_id,
            'last_accessed_at': utils.datetime_to_iso8601(owner_vo.last_accessed_at),
            'created_at': utils.datetime_to_iso8601(owner_vo.created_at)

        })

    return domain_owner_pb2.DomainOwnerInfo(**info)



import functools
from spaceone.core.pygrpc.message_type import *
from spaceone.api.identity.v1 import user_pb2

__all__ = ['FindUserInfo', 'FindUsersInfo']


def FindUserInfo(user_data: dict, minimal=False):
    info = {
        'user_id': user_data.get('user_id'),
        'name': user_data.get('name'),
        'email': user_data.get('email')
    }
    if not minimal:
        tags = {}
        for tag in user_data.get('tags', []):
            tags[tag['key']] = tag['value']

        info.update({
            'tags': change_struct_type(tags),
        })

    return user_pb2.FindUserInfo(**info)


def FindUsersInfo(find_users_vos, total_count, **kwargs):
    results = list(map(functools.partial(FindUserInfo, **kwargs), find_users_vos))
    return user_pb2.FindUsersInfo(results=results, total_count=total_count)



import pytz
from spaceone.core.service import *
from spaceone.core.error import *
from spaceone.identity.manager import DomainOwnerManager


@authentication_handler(exclude=['create'])
@authorization_handler(exclude=['create'])
@mutation_handler(exclude=['create'])
@event_handler
class DomainOwnerService(BaseService):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.domain_owner_mgr: DomainOwnerManager = self.locator.get_manager('DomainOwnerManager')

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['owner_id', 'password', 'domain_id'])
    def create(self, params):
        """ Create domain owner

        Args:
            params (dict): {
                'owner_id': 'str',
                'password': 'str',
                'name': 'str',
                'email': 'str',
                'language': 'str',
                'timezone': 'str',
                'domain_id': 'str'
            }

        Returns:
            domain_owner_vo (object)
        """

        if 'timezone' in params:
            self._check_timezone(params['timezone'])

        return self.domain_owner_mgr.create_owner(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['owner_id', 'domain_id'])
    def update(self, params):
        """ Update domain owner

        Args:
            params (dict): {
                'owner_id': 'str',
                'password': 'str',
                'name': 'str',
                'email': 'str',
                'language': 'str',
                'timezone': 'str',
                'domain_id': 'str'
            }

        Returns:
            domain_owner_vo (object)
        """

        if 'timezone' in params:
            self._check_timezone(params['timezone'])

        return self.domain_owner_mgr.update_owner(params)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id', 'owner_id'])
    def delete(self, params):
        """ Delete domain owner

        Args:
            params (dict): {
                'owner_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """
        self.domain_owner_mgr.delete_owner(params['domain_id'], params['owner_id'])

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    def get(self, params):
        """ Delete domain owner

        Args:
            params (dict): {
                'owner_id': 'str',
                'domain_id': 'str',
                'only': 'list'
            }

        Returns:
            domain_owner_vo (object)
        """
        return self.domain_owner_mgr.get_owner(params['domain_id'], params.get('owner_id'), params.get('only'))

    @staticmethod
    def _check_timezone(timezone):
        if timezone not in pytz.all_timezones:
            raise ERROR_INVALID_PARAMETER(key='timezone', reason='Timezone is invalid.')
